from typing import TYPE_CHECKING, Any, Optional

from great_expectations.core.usage_statistics.anonymizers.base import BaseAnonymizer

if TYPE_CHECKING:
    from great_expectations.core.usage_statistics.anonymizers.anonymizer import (
        Anonymizer,
    )


class ActionAnonymizer(BaseAnonymizer):
    def __init__(
        self,
        aggregate_anonymizer: "Anonymizer",
        salt: Optional[str] = None,
    ) -> None:
        super().__init__(salt=salt)

        self._aggregate_anonymizer = aggregate_anonymizer

    def anonymize(
        self,
        action_name: str,
        action_obj: Optional[object] = None,
        action_config: Optional[dict] = None,
        obj: Optional[object] = None,
    ) -> Any:
        anonymized_info_dict: dict = {
            "anonymized_name": self._anonymize_string(action_name),
        }

        self._anonymize_object_info(
            object_=action_obj,
            object_config=action_config,
            anonymized_info_dict=anonymized_info_dict,
            runtime_environment={"module_name": "great_expectations.checkpoint"},
        )

        return anonymized_info_dict

    def can_handle(self, obj: Optional[object] = None, **kwargs) -> bool:
        return "action_name" in kwargs



from typing import List

import numpy as np
import pandas as pd

from tests.integration.db.taxi_data_utils import (
    _execute_taxi_splitting_test_cases,
    _get_loaded_table,
    _is_dialect_athena,
)
from tests.integration.fixtures.split_and_sample_data.splitter_test_cases_and_fixtures import (
    TaxiSplittingTestCasesBase,
    TaxiSplittingTestCasesMultiColumnValues,
    TaxiTestData,
)
from tests.test_utils import LoadedTable, get_connection_string_and_dialect

if __name__ == "test_script_module":

    dialect: str
    connection_string: str
    dialect, connection_string = get_connection_string_and_dialect(
        athena_db_name_env_var="ATHENA_TEN_TRIPS_DB_NAME"
    )
    print(f"Testing dialect: {dialect}")

    loaded_table: LoadedTable = _get_loaded_table(dialect=dialect)

    table_name: str = loaded_table.table_name
    test_df: pd.DataFrame = loaded_table.inserted_dataframe

    test_column_names: List[str] = [
        "rate_code_id",
        "payment_type",
    ]

    if _is_dialect_athena(dialect):
        test_column_name: str
        df_null: pd.DataFrame
        df_nonnull: pd.DataFrame
        for test_column_name in test_column_names:
            df_null = test_df[test_df[test_column_name].isnull()]
            df_null[test_column_name] = df_null[test_column_name].apply(
                lambda x: None if np.isnan(x) else x
            )
            df_nonnull = test_df[~test_df[test_column_name].isnull()]
            df_nonnull[test_column_name] = df_nonnull[test_column_name].astype(int)
            test_df = pd.concat([df_null, df_nonnull])

    taxi_test_data: TaxiTestData = TaxiTestData(
        test_df=test_df,
        test_column_name=None,
        test_column_names=test_column_names,
    )
    taxi_splitting_test_cases: TaxiSplittingTestCasesBase = (
        TaxiSplittingTestCasesMultiColumnValues(taxi_test_data=taxi_test_data)
    )
    _execute_taxi_splitting_test_cases(
        taxi_splitting_test_cases=taxi_splitting_test_cases,
        connection_string=connection_string,
        table_name=table_name,
    )



from great_expectations.util import verify_dynamic_loading_support

from .store import Store  # isort:skip
from .store_backend import (  # isort:skip
    StoreBackend,
)
from .gx_cloud_store_backend import GXCloudStoreBackend  # isort:skip
from .ge_cloud_store_backend import GeCloudStoreBackend  # isort:skip
from .tuple_store_backend import (  # isort:skip
    TupleFilesystemStoreBackend,
    TupleGCSStoreBackend,
    TupleS3StoreBackend,
    TupleStoreBackend,
    TupleAzureBlobStoreBackend,
)
from .database_store_backend import DatabaseStoreBackend  # isort:skip
from .inline_store_backend import InlineStoreBackend  # isort:skip
from .in_memory_store_backend import InMemoryStoreBackend  # isort:skip
from .configuration_store import ConfigurationStore  # isort:skip
from .checkpoint_store import CheckpointStore  # isort:skip
from .metric_store import (  # isort:skip
    EvaluationParameterStore,
    MetricStore,
)
from .expectations_store import ExpectationsStore  # isort:skip
from .validations_store import ValidationsStore  # isort:skip
from .query_store import SqlAlchemyQueryStore  # isort:skip
from .html_site_store import HtmlSiteStore  # isort:skip
from .datasource_store import DatasourceStore  # isort:skip
from .profiler_store import ProfilerStore  # isort:skip
from .data_context_store import DataContextStore  # isort:skip



import uuid

import pytest

from great_expectations.core.batch import BatchRequest
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.usage_statistics.anonymizers.anonymizer import Anonymizer


@pytest.fixture
def anonymizer_with_consistent_salt() -> Anonymizer:
    anonymizer = Anonymizer(salt="00000000-0000-0000-0000-00000000a004")
    return anonymizer


# The following empty classes are used in this test module only.
# They are used to ensure class hierarchy is appropriately processed by Anonymizer utilities.


class BaseTestClass:
    pass


class TestClass(BaseTestClass):
    pass


class MyCustomTestClass(TestClass):
    pass


class SomeOtherClass:
    pass


class MyCustomExpectationSuite(ExpectationSuite):
    pass


class MyCustomMultipleInheritanceClass(ExpectationSuite, BatchRequest):
    pass


def test_anonymizer_no_salt():
    # No salt will generate a random one.
    anonymizer1 = Anonymizer()
    anonymizer2 = Anonymizer()

    test_name = "i_am_a_name"

    anon_name_1 = anonymizer1._anonymize_string(test_name)
    anon_name_2 = anonymizer2._anonymize_string(test_name)
    assert anon_name_1 != anon_name_2
    assert len(anon_name_1) == 32
    assert len(anon_name_2) == 32

    # Provided different salt will produce different results
    anonymizer1 = Anonymizer("hello, friend")
    anonymizer2 = Anonymizer("hello, enemy")

    test_name = "i_am_a_name"

    anon_name_1 = anonymizer1._anonymize_string(test_name)
    anon_name_2 = anonymizer2._anonymize_string(test_name)
    assert anon_name_1 != anon_name_2
    assert len(anon_name_1) == 32
    assert len(anon_name_2) == 32


def test_anonymizer_consistent_salt():
    # Provided same salt will produce same results
    data_context_id = str(uuid.uuid4())
    anonymizer1 = Anonymizer(data_context_id)
    anonymizer2 = Anonymizer(data_context_id)

    test_name = "i_am_a_name"

    anon_name_1 = anonymizer1._anonymize_string(test_name)
    anon_name_2 = anonymizer2._anonymize_string(test_name)
    assert anon_name_1 == anon_name_2
    assert len(anon_name_1) == 32
    assert len(anon_name_2) == 32


def test_anonymizer_get_parent_class():
    """
    What does this test and why?
    The method Anonymizer.get_parent_class() should return the name of the parent class if it is or is a subclass of one of the classes_to_check. If not, it should return None. It should do so regardless of the parameter used to pass in the object definition (object_, object_class, object_config). It should also return the first matching class in classes_to_check, even if a later class also matches.
    """
    anonymizer = Anonymizer()

    # classes_to_check in order of inheritance hierarchy
    classes_to_check = [TestClass, BaseTestClass]
    assert (
        anonymizer.get_parent_class(
            classes_to_check=classes_to_check, object_class=MyCustomTestClass
        )
        == "TestClass"
    )
    assert (
        anonymizer.get_parent_class(
            classes_to_check=classes_to_check, object_class=SomeOtherClass
        )
        is None
    )
    classes_to_check = [BaseTestClass]
    assert (
        anonymizer.get_parent_class(
            classes_to_check=classes_to_check, object_class=TestClass
        )
        == "BaseTestClass"
    )

    # classes_to_check in order of inheritance hierarchy
    my_custom_test_class = MyCustomTestClass()
    test_class = TestClass()
    some_other_class = SomeOtherClass()
    classes_to_check = [TestClass, BaseTestClass]
    assert (
        anonymizer.get_parent_class(
            classes_to_check=classes_to_check, object_=my_custom_test_class
        )
        == "TestClass"
    )
    assert (
        anonymizer.get_parent_class(
            classes_to_check=classes_to_check, object_=some_other_class
        )
        is None
    )
    classes_to_check = [BaseTestClass]
    assert (
        anonymizer.get_parent_class(
            classes_to_check=classes_to_check, object_=test_class
        )
        == "BaseTestClass"
    )

    # classes_to_check in order of inheritance hierarchy
    my_custom_test_class_config = {
        "class_name": "MyCustomTestClass",
        "module_name": "tests.core.usage_statistics.test_anonymizer",
    }
    test_class_config = {
        "class_name": "TestClass",
        "module_name": "tests.core.usage_statistics.test_anonymizer",
    }
    some_other_class_config = {
        "class_name": "SomeOtherClass",
        "module_name": "tests.core.usage_statistics.test_anonymizer",
    }
    classes_to_check = [TestClass, BaseTestClass]
    assert (
        anonymizer.get_parent_class(
            classes_to_check=classes_to_check, object_config=my_custom_test_class_config
        )
        == "TestClass"
    )
    assert (
        anonymizer.get_parent_class(
            classes_to_check=classes_to_check, object_config=some_other_class_config
        )
        is None
    )
    classes_to_check = [BaseTestClass]
    assert (
        anonymizer.get_parent_class(
            classes_to_check=classes_to_check, object_config=test_class_config
        )
        == "BaseTestClass"
    )


def test_anonymize_object_info_with_core_ge_object(
    anonymizer_with_consistent_salt: Anonymizer,
):
    anonymized_result: dict = anonymizer_with_consistent_salt._anonymize_object_info(
        anonymized_info_dict={},
        object_=ExpectationSuite(expectation_suite_name="my_suite"),
    )

    assert anonymized_result == {"parent_class": "ExpectationSuite"}


def test_anonymize_object_info_with_custom_user_defined_object_with_single_parent(
    anonymizer_with_consistent_salt: Anonymizer,
):
    anonymized_result: dict = anonymizer_with_consistent_salt._anonymize_object_info(
        anonymized_info_dict={},
        object_=MyCustomExpectationSuite(expectation_suite_name="my_suite"),
    )

    assert anonymized_result == {
        "anonymized_class": "54ab2657b855f8075e5d1f28e81ca7cd",
        "parent_class": "ExpectationSuite",
    }


def test_anonymize_object_info_with_custom_user_defined_object_with_no_parent(
    anonymizer_with_consistent_salt: Anonymizer,
):
    anonymized_result: dict = anonymizer_with_consistent_salt._anonymize_object_info(
        anonymized_info_dict={}, object_=BaseTestClass()
    )

    assert anonymized_result == {
        "anonymized_class": "760bfe8b56356bcd56012edfd512019b",
        "parent_class": "__not_recognized__",
    }


def test_anonymize_object_info_with_custom_user_defined_object_with_multiple_parents(
    anonymizer_with_consistent_salt: Anonymizer,
):
    anonymized_result: dict = anonymizer_with_consistent_salt._anonymize_object_info(
        anonymized_info_dict={},
        object_=MyCustomMultipleInheritanceClass(expectation_suite_name="my_name"),
    )

    assert anonymized_result == {
        "anonymized_class": "1e1716661acfa73d538a191ed13efcfd",
        "parent_class": "ExpectationSuite,BatchRequest",
    }


def test_anonymize_object_info_with_missing_args_raises_error(
    anonymizer_with_consistent_salt: Anonymizer,
):
    with pytest.raises(AssertionError) as e:
        anonymizer_with_consistent_salt._anonymize_object_info(
            anonymized_info_dict={},
            object_=None,
            object_class=None,
            object_config=None,
        )

    assert "Must pass either" in str(e.value)



from __future__ import annotations

import contextlib
import enum
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Generator, Optional

from great_expectations.core.config_provider import (
    _ConfigurationProvider,  # noqa: TCH001
)
from great_expectations.core.data_context_key import DataContextKey  # noqa: TCH001
from great_expectations.data_context.types.base import (
    AnonymizedUsageStatisticsConfig,  # noqa: TCH001
    ConcurrencyConfig,  # noqa: TCH001
    DataContextConfig,  # noqa: TCH001
    IncludeRenderedContentConfig,  # noqa: TCH001
    NotebookConfig,  # noqa: TCH001
    ProgressBarsConfig,  # noqa: TCH001
)
from great_expectations.data_context.types.resource_identifiers import (
    ConfigurationIdentifier,
    GXCloudIdentifier,
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.file_data_context import (
        FileDataContext,
    )
    from great_expectations.data_context.store import DataContextStore
    from great_expectations.experimental.datasources.interfaces import (
        Datasource as XDatasource,
    )

logger = logging.getLogger(__file__)


class DataContextVariableSchema(str, enum.Enum):
    ALL_VARIABLES = (
        "data_context_variables"  # If retrieving/setting the entire config at once
    )
    CONFIG_VERSION = "config_version"
    DATASOURCES = "datasources"
    EXPECTATIONS_STORE_NAME = "expectations_store_name"
    VALIDATIONS_STORE_NAME = "validations_store_name"
    EVALUATION_PARAMETER_STORE_NAME = "evaluation_parameter_store_name"
    CHECKPOINT_STORE_NAME = "checkpoint_store_name"
    PROFILER_STORE_NAME = "profiler_store_name"
    PLUGINS_DIRECTORY = "plugins_directory"
    VALIDATION_OPERATORS = "validation_operators"
    STORES = "stores"
    DATA_DOCS_SITES = "data_docs_sites"
    NOTEBOOKS = "notebooks"
    CONFIG_VARIABLES_FILE_PATH = "config_variables_file_path"
    ANONYMOUS_USAGE_STATISTICS = "anonymous_usage_statistics"
    CONCURRENCY = "concurrency"
    PROGRESS_BARS = "progress_bars"
    INCLUDE_RENDERED_CONTENT = "include_rendered_content"

    @classmethod
    def has_value(cls, value: str) -> bool:
        """
        Checks whether or not a string is a value from the possible enum pairs.
        """
        return value in cls._value2member_map_


@dataclass
class DataContextVariables(ABC):
    """
    Wrapper object around data context variables set in the `great_expectations.yml` config file.

    Child classes should instantiate their own stores to ensure that changes made to this object
    are persisted for future usage (i.e. filesystem I/O or HTTP request to a Cloud endpoint).

    Should maintain parity with the `DataContextConfig`.

    Args:
        config:          A reference to the DataContextConfig to perform CRUD on.
        config_provider: Responsible for determining config values and substituting them in GET calls.
        _store:          An instance of a DataContextStore with the appropriate backend to persist config changes.
    """

    config: DataContextConfig
    config_provider: _ConfigurationProvider
    _store: Optional[DataContextStore] = None

    def __str__(self) -> str:
        return str(self.config)

    def __repr__(self) -> str:
        return repr(self.config)

    @property
    def store(self) -> DataContextStore:
        if self._store is None:
            self._store = self._init_store()
        return self._store

    @abstractmethod
    def _init_store(self) -> DataContextStore:
        raise NotImplementedError

    def get_key(self) -> DataContextKey:
        """
        Generates the appropriate Store key to retrieve/store configs.
        """
        key = ConfigurationIdentifier(
            configuration_key=DataContextVariableSchema.ALL_VARIABLES
        )
        return key

    def _set(self, attr: DataContextVariableSchema, value: Any) -> None:
        key: str = attr.value
        self.config[key] = value

    def _get(self, attr: DataContextVariableSchema) -> Any:
        key: str = attr.value
        val: Any = self.config[key]
        substituted_val: Any = self.config_provider.substitute_config(val)
        return substituted_val

    def save_config(self) -> Any:
        """
        Persist any changes made to variables utilizing the configured Store.
        """
        key: ConfigurationIdentifier = self.get_key()  # type: ignore[assignment]
        return self.store.set(key=key, value=self.config)

    @property
    def config_version(self) -> Optional[float]:
        return self._get(DataContextVariableSchema.CONFIG_VERSION)

    @config_version.setter
    def config_version(self, config_version: float) -> None:
        self._set(DataContextVariableSchema.CONFIG_VERSION, config_version)

    @property
    def config_variables_file_path(self) -> Optional[str]:
        return self._get(DataContextVariableSchema.CONFIG_VARIABLES_FILE_PATH)

    @config_variables_file_path.setter
    def config_variables_file_path(self, config_variables_file_path: str) -> None:
        self._set(
            DataContextVariableSchema.CONFIG_VARIABLES_FILE_PATH,
            config_variables_file_path,
        )

    @property
    def plugins_directory(self) -> Optional[str]:
        return self._get(DataContextVariableSchema.PLUGINS_DIRECTORY)

    @plugins_directory.setter
    def plugins_directory(self, plugins_directory: str) -> None:
        self._set(DataContextVariableSchema.PLUGINS_DIRECTORY, plugins_directory)

    @property
    def validation_operators(self) -> Optional[dict]:
        return self._get(DataContextVariableSchema.VALIDATION_OPERATORS)

    @validation_operators.setter
    def validation_operators(self, validation_operators: dict) -> None:
        self._set(DataContextVariableSchema.VALIDATION_OPERATORS, validation_operators)

    @property
    def expectations_store_name(self) -> Optional[str]:
        return self._get(DataContextVariableSchema.EXPECTATIONS_STORE_NAME)

    @expectations_store_name.setter
    def expectations_store_name(self, expectations_store_name: str) -> None:
        self._set(
            DataContextVariableSchema.EXPECTATIONS_STORE_NAME, expectations_store_name
        )

    @property
    def validations_store_name(self) -> Optional[str]:
        return self._get(DataContextVariableSchema.VALIDATIONS_STORE_NAME)

    @validations_store_name.setter
    def validations_store_name(self, validations_store_name: str) -> None:
        self._set(
            DataContextVariableSchema.VALIDATIONS_STORE_NAME, validations_store_name
        )

    @property
    def evaluation_parameter_store_name(self) -> Optional[str]:
        return self._get(DataContextVariableSchema.EVALUATION_PARAMETER_STORE_NAME)

    @evaluation_parameter_store_name.setter
    def evaluation_parameter_store_name(
        self, evaluation_parameter_store_name: str
    ) -> None:
        self._set(
            DataContextVariableSchema.EVALUATION_PARAMETER_STORE_NAME,
            evaluation_parameter_store_name,
        )

    @property
    def checkpoint_store_name(self) -> Optional[str]:
        return self._get(DataContextVariableSchema.CHECKPOINT_STORE_NAME)

    @checkpoint_store_name.setter
    def checkpoint_store_name(self, checkpoint_store_name: str) -> None:
        self._set(
            DataContextVariableSchema.CHECKPOINT_STORE_NAME,
            checkpoint_store_name,
        )

    @property
    def profiler_store_name(self) -> Optional[str]:
        return self._get(DataContextVariableSchema.PROFILER_STORE_NAME)

    @profiler_store_name.setter
    def profiler_store_name(self, profiler_store_name: str) -> None:
        self._set(
            DataContextVariableSchema.PROFILER_STORE_NAME,
            profiler_store_name,
        )

    @property
    def stores(self) -> Optional[dict]:
        return self._get(DataContextVariableSchema.STORES)

    @stores.setter
    def stores(self, stores: dict) -> None:
        self._set(DataContextVariableSchema.STORES, stores)

    @property
    def data_docs_sites(self) -> Optional[dict]:
        return self._get(DataContextVariableSchema.DATA_DOCS_SITES)

    @data_docs_sites.setter
    def data_docs_sites(self, data_docs_sites: dict) -> None:
        self._set(DataContextVariableSchema.DATA_DOCS_SITES, data_docs_sites)

    @property
    def anonymous_usage_statistics(
        self,
    ) -> Optional[AnonymizedUsageStatisticsConfig]:
        return self._get(DataContextVariableSchema.ANONYMOUS_USAGE_STATISTICS)

    @anonymous_usage_statistics.setter
    def anonymous_usage_statistics(
        self, anonymous_usage_statistics: AnonymizedUsageStatisticsConfig
    ) -> None:
        self._set(
            DataContextVariableSchema.ANONYMOUS_USAGE_STATISTICS,
            anonymous_usage_statistics,
        )

    @property
    def notebooks(self) -> Optional[NotebookConfig]:
        return self._get(DataContextVariableSchema.NOTEBOOKS)

    @notebooks.setter
    def notebooks(self, notebooks: NotebookConfig) -> None:
        self._set(
            DataContextVariableSchema.NOTEBOOKS,
            notebooks,
        )

    @property
    def concurrency(self) -> Optional[ConcurrencyConfig]:
        return self._get(DataContextVariableSchema.CONCURRENCY)

    @concurrency.setter
    def concurrency(self, concurrency: ConcurrencyConfig) -> None:
        self._set(
            DataContextVariableSchema.CONCURRENCY,
            concurrency,
        )

    @property
    def progress_bars(self) -> Optional[ProgressBarsConfig]:
        return self._get(DataContextVariableSchema.PROGRESS_BARS)

    @progress_bars.setter
    def progress_bars(self, progress_bars: ProgressBarsConfig) -> None:
        self._set(
            DataContextVariableSchema.PROGRESS_BARS,
            progress_bars,
        )

    @property
    def include_rendered_content(self) -> IncludeRenderedContentConfig:
        return self._get(DataContextVariableSchema.INCLUDE_RENDERED_CONTENT)

    @include_rendered_content.setter
    def include_rendered_content(
        self, include_rendered_content: IncludeRenderedContentConfig
    ) -> None:
        self._set(
            DataContextVariableSchema.INCLUDE_RENDERED_CONTENT,
            include_rendered_content,
        )


@dataclass(repr=False)
class EphemeralDataContextVariables(DataContextVariables):
    def _init_store(self) -> DataContextStore:
        from great_expectations.data_context.store.data_context_store import (
            DataContextStore,
        )

        store = DataContextStore(
            store_name="ephemeral_data_context_store",
            store_backend=None,  # Defaults to InMemoryStoreBackend
            runtime_environment=None,
        )
        return store


@dataclass(repr=False)
class FileDataContextVariables(DataContextVariables):
    data_context: FileDataContext = None  # type: ignore[assignment] # post_init ensures field always set

    def __post_init__(self) -> None:
        # Chetan - 20220607 - Although the above argument is not truly optional, we are
        # required to use default values because the parent class defines arguments with default values
        # ("Fields without default values cannot appear after fields with default values").
        #
        # Python 3.10 resolves this issue around dataclass inheritance using `kw_only=True` (https://docs.python.org/3/library/dataclasses.html)
        # This should be modified once our lowest supported version is 3.10.

        if self.data_context is None:
            raise ValueError(
                f"A reference to a data context is required for {self.__class__.__name__}"
            )

    def _init_store(self) -> DataContextStore:
        from great_expectations.data_context.store.data_context_store import (
            DataContextStore,
        )

        store_backend: dict = {
            "class_name": "InlineStoreBackend",
            "resource_type": DataContextVariableSchema.ALL_VARIABLES,
            "data_context": self.data_context,
        }
        store = DataContextStore(
            store_name="file_data_context_store",
            store_backend=store_backend,
            runtime_environment=None,
        )
        return store

    def save_config(self) -> Any:
        """
        Persist any changes made to variables utilizing the configured Store.
        """
        # overridden in order to prevent calling `instantiate_class_from_config` on ZEP objects
        # parent class does not have access to the `data_context`
        with self._zep_objects_stash():
            save_result = super().save_config()
        return save_result

    @contextlib.contextmanager
    def _zep_objects_stash(
        self: FileDataContextVariables,
    ) -> Generator[None, None, None]:
        """
        Temporarily remove and stash zep objects from the datacontext.
        Replace them once the with block ends.

        NOTE: This could be generalized into a stand-alone context manager function,
        but it would need to take in the data_context containing the zep objects.
        """
        config_xdatasources_stash: Dict[
            str, XDatasource
        ] = self.data_context._synchronize_zep_datasources()
        try:
            if config_xdatasources_stash:
                logger.info(
                    f"Stashing `XDatasource` during {type(self).__name__}.save_config() - {len(config_xdatasources_stash)} stashed"
                )
                for xdatasource_name in config_xdatasources_stash.keys():
                    self.data_context.datasources.pop(xdatasource_name)
                # this would be `deep_copy'ed in `instantiate_class_from_config` too
                self.data_context.zep_config.xdatasources = {}
            yield
        except Exception:
            raise
        finally:
            if config_xdatasources_stash:
                logger.info(
                    f"Replacing {len(config_xdatasources_stash)} stashed `XDatasource`s"
                )
                self.data_context.datasources.update(config_xdatasources_stash)
                self.data_context.zep_config.xdatasources = config_xdatasources_stash


@dataclass(repr=False)
class CloudDataContextVariables(DataContextVariables):
    ge_cloud_base_url: str = None  # type: ignore[assignment] # post_init ensures field always set
    ge_cloud_organization_id: str = None  # type: ignore[assignment] # post_init ensures field always set
    ge_cloud_access_token: str = None  # type: ignore[assignment] # post_init ensures field always set

    def __post_init__(self) -> None:
        # Chetan - 20220607 - Although the above arguments are not truly optional, we are
        # required to use default values because the parent class defines arguments with default values
        # ("Fields without default values cannot appear after fields with default values").
        #
        # Python 3.10 resolves this issue around dataclass inheritance using `kw_only=True` (https://docs.python.org/3/library/dataclasses.html)
        # This should be modified once our lowest supported version is 3.10.

        if any(
            attr is None
            for attr in (
                self.ge_cloud_base_url,
                self.ge_cloud_organization_id,
                self.ge_cloud_access_token,
            )
        ):
            raise ValueError(
                f"All of the following attributes are required for{ self.__class__.__name__}:\n  self.ge_cloud_base_url\n  self.ge_cloud_organization_id\n  self.ge_cloud_access_token"
            )

    def _init_store(self) -> DataContextStore:
        from great_expectations.data_context.cloud_constants import GXCloudRESTResource
        from great_expectations.data_context.store.data_context_store import (
            DataContextStore,
        )
        from great_expectations.data_context.store.gx_cloud_store_backend import (
            GXCloudStoreBackend,
        )

        store_backend: dict = {
            "class_name": GXCloudStoreBackend.__name__,
            "ge_cloud_base_url": self.ge_cloud_base_url,
            "ge_cloud_resource_type": GXCloudRESTResource.DATA_CONTEXT_VARIABLES,
            "ge_cloud_credentials": {
                "access_token": self.ge_cloud_access_token,
                "organization_id": self.ge_cloud_organization_id,
            },
            "suppress_store_backend_id": True,
        }
        store = DataContextStore(
            store_name="cloud_data_context_store",
            store_backend=store_backend,
            runtime_environment=None,
        )
        return store

    def get_key(self) -> GXCloudIdentifier:
        """
        Generates a GX Cloud-specific key for use with Stores. See parent "DataContextVariables.get_key" for more details.
        """
        from great_expectations.data_context.cloud_constants import GXCloudRESTResource

        key = GXCloudIdentifier(
            resource_type=GXCloudRESTResource.DATA_CONTEXT_VARIABLES
        )
        return key



import logging
import re
from typing import List, Optional

from great_expectations.core._docs_decorators import public_api
from great_expectations.core.batch import BatchDefinition  # noqa: TCH001
from great_expectations.core.batch_spec import AzureBatchSpec, PathBatchSpec
from great_expectations.datasource.data_connector.inferred_asset_file_path_data_connector import (
    InferredAssetFilePathDataConnector,
)
from great_expectations.datasource.data_connector.util import (
    list_azure_keys,
    sanitize_prefix,
)
from great_expectations.execution_engine import ExecutionEngine  # noqa: TCH001

logger = logging.getLogger(__name__)

try:
    from azure.storage.blob import BlobServiceClient
except ImportError:
    BlobServiceClient = None
    logger.debug(
        "Unable to load BlobServiceClient connection object; install optional Azure Storage Blob dependency for support"
    )


@public_api
class InferredAssetAzureDataConnector(InferredAssetFilePathDataConnector):
    """An Inferred Asset Data Connector used to connect to Azure Blob Storage.

    This Data Connector uses regular expressions to traverse through Azure Blob Storage containers and implicitly
    determine Data Asset name.

    Much of the interaction is performed using a BlobServiceClient. Please refer to
    the `official Azure documentation <https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blobserviceclient?view=azure-python>`_ for
    more information.

    Args:
        name: The name of the Data Connector.
        datasource_name: The name of this Data Connector's Datasource.
        container: The container for Azure Blob Storage.
        execution_engine: The Execution Engine object to used by this Data Connector to read the data.
        default_regex: A regex configuration for filtering data references. The dict can include a regex `pattern` and
            a list of `group_names` for capture groups.
        sorters: A list of sorters for sorting data references.
        name_starts_with: Infer as Data Assets only blobs that begin with this prefix.
        delimiter: When included, will remove any prefix up to the delimiter from the inferred Data Asset names.
        azure_options: Options passed to the `BlobServiceClient`.
        batch_spec_passthrough: Dictionary with keys that will be added directly to the batch spec.
        id: The unique identifier for this Data Connector used when running in cloud mode.
    """

    def __init__(
        self,
        name: str,
        datasource_name: str,
        container: str,
        execution_engine: Optional[ExecutionEngine] = None,
        default_regex: Optional[dict] = None,
        sorters: Optional[list] = None,
        name_starts_with: str = "",
        delimiter: str = "/",
        azure_options: Optional[dict] = None,
        batch_spec_passthrough: Optional[dict] = None,
        id: Optional[str] = None,
    ) -> None:
        logger.debug(f'Constructing InferredAssetAzureDataConnector "{name}".')

        super().__init__(
            name=name,
            id=id,
            datasource_name=datasource_name,
            execution_engine=execution_engine,
            default_regex=default_regex,
            sorters=sorters,
            batch_spec_passthrough=batch_spec_passthrough,
        )

        self._container = container
        self._name_starts_with = sanitize_prefix(name_starts_with)
        self._delimiter = delimiter

        if azure_options is None:
            azure_options = {}

        # Thanks to schema validation, we are guaranteed to have one of `conn_str` or `account_url` to
        # use in authentication (but not both). If the format or content of the provided keys is invalid,
        # the assignment of `self._account_name` and `self._azure` will fail and an error will be raised.
        conn_str: Optional[str] = azure_options.get("conn_str")
        account_url: Optional[str] = azure_options.get("account_url")
        assert bool(conn_str) ^ bool(
            account_url
        ), "You must provide one of `conn_str` or `account_url` to the `azure_options` key in your config (but not both)"

        try:
            if conn_str is not None:
                self._account_name = re.search(  # type: ignore[union-attr]
                    r".*?AccountName=(.+?);.*?", conn_str
                ).group(1)
                self._azure = BlobServiceClient.from_connection_string(**azure_options)
            elif account_url is not None:
                self._account_name = re.search(  # type: ignore[union-attr]
                    r"(?:https?://)?(.+?).blob.core.windows.net", account_url
                ).group(1)
                self._azure = BlobServiceClient(**azure_options)
        except (TypeError, AttributeError):
            raise ImportError(
                "Unable to load Azure BlobServiceClient (it is required for InferredAssetAzureDataConnector). \
                Please ensure that you have provided the appropriate keys to `azure_options` for authentication."
            )

    def build_batch_spec(self, batch_definition: BatchDefinition) -> AzureBatchSpec:
        """
        Build BatchSpec from batch_definition by calling DataConnector's build_batch_spec function.

        Args:
            batch_definition (BatchDefinition): to be used to build batch_spec

        Returns:
            BatchSpec built from batch_definition
        """
        batch_spec: PathBatchSpec = super().build_batch_spec(
            batch_definition=batch_definition
        )
        return AzureBatchSpec(batch_spec)

    def _get_data_reference_list(
        self, data_asset_name: Optional[str] = None
    ) -> List[str]:
        """
        List objects in the underlying data store to create a list of data_references.

        This method is used to refresh the cache.
        """
        query_options: dict = {
            "container": self._container,
            "name_starts_with": self._name_starts_with,
            "delimiter": self._delimiter,
        }

        path_list: List[str] = list_azure_keys(
            azure=self._azure,
            query_options=query_options,
            recursive=True,
        )
        return path_list

    def _get_full_file_path(
        self,
        path: str,
        data_asset_name: Optional[str] = None,
    ) -> str:
        # data_asset_name isn't used in this method.
        # It's only kept for compatibility with parent methods.
        template_arguments: dict = {
            "account_name": self._account_name,
            "container": self._container,
            "path": path,
        }
        return self.resolve_data_reference(template_arguments=template_arguments)



import pandas as pd

from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.column_aggregate_metric_provider import (
    ColumnAggregateMetricProvider,
    column_aggregate_partial,
    column_aggregate_value,
)
from great_expectations.expectations.metrics.import_manager import F, sa


class ColumnValuesLengthMin(ColumnAggregateMetricProvider):
    metric_name = "column_values.length.min"

    @column_aggregate_value(engine=PandasExecutionEngine, filter_column_isnull=True)
    def _pandas(cls, column: pd.Series, **kwargs: dict) -> int:
        return column.map(len).min()

    @column_aggregate_partial(
        engine=SqlAlchemyExecutionEngine, filter_column_isnull=True
    )
    def _sqlalchemy(cls, column, **kwargs: dict):
        return sa.func.min(sa.func.length(column))

    @column_aggregate_partial(engine=SparkDFExecutionEngine, filter_column_isnull=True)
    def _spark(cls, column, **kwargs: dict):
        return F.min(F.length(column))



import pandas as pd
from ruamel import yaml

import great_expectations as gx
from great_expectations.core.batch import RuntimeBatchRequest

context = gx.get_context()

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/in_memory/pandas_python_example.py datasource_config">
datasource_config = {
    "name": "example_datasource",
    "class_name": "Datasource",
    "module_name": "great_expectations.datasource",
    "execution_engine": {
        "module_name": "great_expectations.execution_engine",
        "class_name": "PandasExecutionEngine",
    },
    "data_connectors": {
        "default_runtime_data_connector_name": {
            "class_name": "RuntimeDataConnector",
            "module_name": "great_expectations.datasource.data_connector",
            "batch_identifiers": ["default_identifier_name"],
        },
    },
}
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/in_memory/pandas_python_example.py test_yaml_config">
context.test_yaml_config(yaml.dump(datasource_config))
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/in_memory/pandas_python_example.py add_datasource">
context.add_datasource(**datasource_config)
# </snippet>

# creating our example Pandas dataframe
df = pd.DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9]], columns=["a", "b", "c"])

# Here is a RuntimeBatchRequest that takes in our df as batch_data
batch_request = RuntimeBatchRequest(
    datasource_name="example_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="<YOUR_MEANINGFUL_NAME>",  # This can be anything that identifies this data_asset for you
    runtime_parameters={"batch_data": df},  # df is your dataframe
    batch_identifiers={"default_identifier_name": "default_identifier"},
)

context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)



from great_expectations.core.expectation_diagnostics.expectation_test_data_cases import (
    ExpectationLegacyTestCaseAdapter,
    TestData,
)


# @pytest.mark.skip(reason="Not yet supported")
def test_basic_instantiation_of_TestData():
    TestData(
        **{
            "a": ["aaa", "abb", "acc", "add", "bee"],
            "b": ["aaa", "abb", "acc", "bdd", None],
            "column_name with space": ["aaa", "abb", "acc", "add", "bee"],
        }
    )


def test_basic_instantiation_of_ExpectationTestCase():
    my_test_case = ExpectationLegacyTestCaseAdapter(
        **{
            "title": "basic_negative_test",
            "exact_match_out": False,
            "in": {"column": "a", "regex": "^a", "mostly": 0.9},
            "out": {
                "success": False,
                "unexpected_index_list": [4],
                "unexpected_list": ["bee"],
            },
            "suppress_test_for": ["sqlite", "mssql"],
        }
    )

    assert my_test_case.keys() == {
        "title",
        "input",
        "output",
        "suppress_test_for",
        "exact_match_out",
        "include_in_gallery",
        "only_for",
    }

    assert my_test_case.input == {"column": "a", "regex": "^a", "mostly": 0.9}
    assert my_test_case.input["column"] == "a"
    # assert my_test_case.input.column == "a" #Not supported yet. We'll need to add types to input and output to get there.



import json
from typing import List, Union

import pytest

from great_expectations.core.batch import Batch, BatchRequest, IDDict
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.execution_engine.sqlalchemy_batch_data import (
    SqlAlchemyBatchData,
)

yaml = YAMLHandler()

try:
    sqlalchemy = pytest.importorskip("sqlalchemy")
except ImportError:
    sqlalchemy = None


def test_get_batch(data_context_with_simple_sql_datasource_for_testing_get_batch):
    context = data_context_with_simple_sql_datasource_for_testing_get_batch

    print(
        json.dumps(
            context.datasources["my_sqlite_db"].get_available_data_asset_names(),
            indent=4,
        )
    )

    # Successful specification using a typed BatchRequest
    with pytest.deprecated_call():
        context.get_batch(
            batch_request=BatchRequest(
                datasource_name="my_sqlite_db",
                data_connector_name="daily",
                data_asset_name="table_partitioned_by_date_column__A",
                data_connector_query=IDDict(
                    batch_filter_parameters={"date": "2020-01-15"}
                ),
            )
        )

        # Failed specification using an untyped BatchRequest
        with pytest.raises(TypeError):
            context.get_batch(
                batch_request={
                    "datasource_name": "my_sqlite_db",
                    "data_connector_name": "daily",
                    "data_asset_name": "table_partitioned_by_date_column__A",
                    "data_connector_query": {
                        "batch_filter_parameters": {"date": "2020-01-15"}
                    },
                }
            )

        # Failed specification using an incomplete BatchRequest
        with pytest.raises(ValueError):
            context.get_batch(
                batch_request=BatchRequest(
                    datasource_name="my_sqlite_db",
                    data_connector_name="daily",
                    data_asset_name="table_partitioned_by_date_column__A",
                    data_connector_query=IDDict(batch_filter_parameters={}),
                )
            )

        # Failed specification using an incomplete BatchRequest
        with pytest.raises(ValueError):
            context.get_batch(
                batch_request=BatchRequest(
                    datasource_name="my_sqlite_db",
                    data_connector_name="daily",
                    data_asset_name="table_partitioned_by_date_column__A",
                )
            )

        # Failed specification using an incomplete BatchRequest
        with pytest.raises(TypeError):
            context.get_batch(
                batch_request=BatchRequest(
                    datasource_name="my_sqlite_db", data_connector_name="daily"
                )
            )

        # Failed specification using an incomplete BatchRequest
        # with pytest.raises(ValueError):
        with pytest.raises(TypeError):
            context.get_batch(
                batch_request=BatchRequest(
                    data_connector_name="daily",
                    data_asset_name="table_partitioned_by_date_column__A",
                    data_connector_query=IDDict(batch_filter_parameters={}),
                )
            )

        # Successful specification using parameters
        context.get_batch(
            datasource_name="my_sqlite_db",
            data_connector_name="daily",
            data_asset_name="table_partitioned_by_date_column__A",
            date="2020-01-15",
        )

        # Successful specification using parameters without parameter names for the identifying triple
        # This is the thinnest this can plausibly get.
        context.get_batch(
            "my_sqlite_db",
            "daily",
            "table_partitioned_by_date_column__A",
            date="2020-01-15",
        )

        # Successful specification using parameters without parameter names for the identifying triple
        # In the case of a data_asset containing a single Batch, we don't even need parameters
        context.get_batch(
            "my_sqlite_db",
            "whole_table",
            "table_partitioned_by_date_column__A",
        )

        # Successful specification using parameters and data_connector_query
        context.get_batch(
            "my_sqlite_db",
            "daily",
            "table_partitioned_by_date_column__A",
            data_connector_query=IDDict(
                {"batch_filter_parameters": {"date": "2020-01-15"}}
            ),
        )

        # Successful specification using parameters and batch_identifiers
        context.get_batch(
            "my_sqlite_db",
            "daily",
            "table_partitioned_by_date_column__A",
            batch_identifiers={"date": "2020-01-15"},
        )


def test_get_validator_bad_batch_request(
    data_context_with_simple_sql_datasource_for_testing_get_batch,
):
    context: "DataContext" = (
        data_context_with_simple_sql_datasource_for_testing_get_batch
    )
    context.add_expectation_suite("my_expectations")
    batch_request: BatchRequest = BatchRequest(
        datasource_name="my_sqlite_db",
        data_connector_name="daily",
        data_asset_name="i_dont_exist",
        data_connector_query=IDDict(batch_filter_parameters={"date": "2020-01-15"}),
    )
    with pytest.raises(KeyError):
        # as a result of introspection, the data_assets will already be loaded into the cache.
        # an incorrect data_asset_name will result in a key error
        context.get_validator(
            batch_request=batch_request, expectation_suite_name="my_expectations"
        )


def test_get_validator(data_context_with_simple_sql_datasource_for_testing_get_batch):
    context = data_context_with_simple_sql_datasource_for_testing_get_batch
    context.add_expectation_suite("my_expectations")

    # Successful specification using a typed BatchRequest
    context.get_validator(
        batch_request=BatchRequest(
            datasource_name="my_sqlite_db",
            data_connector_name="daily",
            data_asset_name="table_partitioned_by_date_column__A",
            data_connector_query=IDDict(batch_filter_parameters={"date": "2020-01-15"}),
        ),
        expectation_suite_name="my_expectations",
    )

    # Failed specification using an untyped BatchRequest
    with pytest.raises(TypeError):
        context.get_validator(
            batch_request={
                "datasource_name": "my_sqlite_db",
                "data_connector_name": "daily",
                "data_asset_name": "table_partitioned_by_date_column__A",
                "data_connector_query": {
                    "batch_filter_parameters": {"date": "2020-01-15"}
                },
            },
            expectation_suite_name="my_expectations",
        )

    # A BatchRequest specified without the date batch_filter_parameter will return all 30 batches.
    assert (
        len(
            context.get_validator(
                batch_request=BatchRequest(
                    datasource_name="my_sqlite_db",
                    data_connector_name="daily",
                    data_asset_name="table_partitioned_by_date_column__A",
                    data_connector_query=IDDict(batch_filter_parameters={}),
                ),
                expectation_suite_name="my_expectations",
            ).batches
        )
        == 34
    )

    # A BatchRequest specified without a data_connector_query will return all 30 batches.
    assert (
        len(
            context.get_validator(
                batch_request=BatchRequest(
                    datasource_name="my_sqlite_db",
                    data_connector_name="daily",
                    data_asset_name="table_partitioned_by_date_column__A",
                ),
                expectation_suite_name="my_expectations",
            ).batches
        )
        == 34
    )

    # Failed specification using an incomplete BatchRequest
    with pytest.raises(TypeError):
        context.get_validator(
            batch_request=BatchRequest(
                datasource_name="my_sqlite_db", data_connector_name="daily"
            ),
            expectation_suite_name="my_expectations",
        )

    # Failed specification using an incomplete BatchRequest
    # with pytest.raises(ValueError):
    with pytest.raises(TypeError):
        context.get_validator(
            batch_request=BatchRequest(
                data_connector_name="daily",
                data_asset_name="table_partitioned_by_date_column__A",
                data_connector_query=IDDict(batch_filter_parameters={}),
            ),
            expectation_suite_name="my_expectations",
        )

    # Successful specification using parameters
    context.get_validator(
        datasource_name="my_sqlite_db",
        data_connector_name="daily",
        data_asset_name="table_partitioned_by_date_column__A",
        expectation_suite_name="my_expectations",
        date="2020-01-15",
    )

    # Successful specification using parameters without parameter names for the identifying triple
    # This is the thinnest this can plausibly get.
    context.get_validator(
        "my_sqlite_db",
        "daily",
        "table_partitioned_by_date_column__A",
        expectation_suite_name="my_expectations",
        date="2020-01-15",
    )

    # Successful specification using parameters without parameter names for the identifying triple
    # In the case of a data_asset containing a single Batch, we don't even need parameters
    context.get_validator(
        "my_sqlite_db",
        "whole_table",
        "table_partitioned_by_date_column__A",
        expectation_suite_name="my_expectations",
    )

    # Successful specification using parameters and data_connector_query
    context.get_validator(
        "my_sqlite_db",
        "daily",
        "table_partitioned_by_date_column__A",
        data_connector_query=IDDict(
            {"batch_filter_parameters": {"date": "2020-01-15"}}
        ),
        expectation_suite_name="my_expectations",
    )

    # Successful specification using parameters and batch_identifiers
    context.get_validator(
        "my_sqlite_db",
        "daily",
        "table_partitioned_by_date_column__A",
        batch_identifiers={"date": "2020-01-15"},
        expectation_suite_name="my_expectations",
    )


def test_get_validator_expectation_suite_options(
    data_context_with_simple_sql_datasource_for_testing_get_batch,
):
    context = data_context_with_simple_sql_datasource_for_testing_get_batch
    context.add_expectation_suite("some_expectations")

    # Successful specification with an existing expectation_suite_name
    context.get_validator(
        datasource_name="my_sqlite_db",
        data_connector_name="daily",
        data_asset_name="table_partitioned_by_date_column__A",
        expectation_suite_name="some_expectations",
        date="2020-01-15",
    )

    # Successful specification with a fetched ExpectationSuite object
    some_expectations = context.get_expectation_suite("some_expectations")
    context.get_validator(
        datasource_name="my_sqlite_db",
        data_connector_name="daily",
        data_asset_name="table_partitioned_by_date_column__A",
        expectation_suite=some_expectations,
        date="2020-01-15",
    )

    # Successful specification with a fresh ExpectationSuite object
    some_more_expectations = context.add_expectation_suite(
        expectation_suite_name="some_more_expectations"
    )
    context.get_validator(
        datasource_name="my_sqlite_db",
        data_connector_name="daily",
        data_asset_name="table_partitioned_by_date_column__A",
        expectation_suite=some_more_expectations,
        date="2020-01-15",
    )

    # Successful specification using overwrite_existing_expectation_suite
    context.get_validator(
        batch_request=BatchRequest(
            datasource_name="my_sqlite_db",
            data_connector_name="daily",
            data_asset_name="table_partitioned_by_date_column__A",
            data_connector_query=IDDict(batch_filter_parameters={"date": "2020-01-15"}),
        ),
        create_expectation_suite_with_name="yet_more_expectations",
    )

    # Failed specification: incorrectly typed expectation suite
    with pytest.raises(TypeError):
        context.get_validator(
            datasource_name="my_sqlite_db",
            data_connector_name="daily",
            data_asset_name="table_partitioned_by_date_column__A",
            expectation_suite={
                "im": "a",
                "dictionary": "not a",
                "ExepctationSuite": False,
            },
            date="2020-01-15",
        )


def test_get_batch_list_from_new_style_datasource_with_sql_datasource(
    sa, data_context_with_simple_sql_datasource_for_testing_get_batch
):
    context = data_context_with_simple_sql_datasource_for_testing_get_batch

    batch_request: Union[dict, BatchRequest] = {
        "datasource_name": "my_sqlite_db",
        "data_connector_name": "daily",
        "data_asset_name": "table_partitioned_by_date_column__A",
        "data_connector_query": {"batch_filter_parameters": {"date": "2020-01-15"}},
    }
    batch_list: List[Batch] = context.get_batch_list(**batch_request)

    assert len(batch_list) == 1

    batch: Batch = batch_list[0]
    assert batch.batch_spec is not None
    assert (
        batch.batch_definition["data_asset_name"]
        == "table_partitioned_by_date_column__A"
    )
    assert batch.batch_definition["batch_identifiers"] == {"date": "2020-01-15"}
    assert isinstance(batch.data, SqlAlchemyBatchData)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import nbformat

from great_expectations.render.renderer.renderer import Renderer
from great_expectations.util import (
    convert_json_string_to_be_python_compliant,
    lint_code,
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )


class BaseNotebookRenderer(Renderer):
    """
    Abstract base class for methods that help with rendering a jupyter notebook.
    """

    def __init__(self, context: Optional[AbstractDataContext] = None) -> None:
        super().__init__()
        self.context = context
        # Add cells to this notebook, then render by implementing a
        # self.render() and/or self.render_to_disk() method(s):
        self._notebook: Optional[nbformat.NotebookNode] = None

    def add_code_cell(
        self, code: str, lint: bool = False, enforce_py_syntax: bool = True
    ) -> None:
        """
        Add the given code as a new code cell.
        Args:
            code: Code to render into the notebook cell
            lint: Whether to lint the code before adding it
            enforce_py_syntax: Directive to convert code to Python-compliant format

        Returns:
            Nothing, adds a cell to the class instance notebook
        """
        if enforce_py_syntax:
            code = convert_json_string_to_be_python_compliant(code)

        if lint:
            code = lint_code(code).rstrip("\n")

        cell = nbformat.v4.new_code_cell(code)
        self._notebook["cells"].append(cell)

    def add_markdown_cell(self, markdown: str) -> None:
        """
        Add the given markdown as a new markdown cell.
        Args:
            markdown: Code to render into the notebook cell

        Returns:
            Nothing, adds a cell to the class instance notebook
        """
        cell = nbformat.v4.new_markdown_cell(markdown)
        self._notebook["cells"].append(cell)

    @classmethod
    def write_notebook_to_disk(
        cls, notebook: nbformat.NotebookNode, notebook_file_path: str
    ) -> None:
        """
        Write a given Jupyter notebook to disk.
        Args:
            notebook: Jupyter notebook
            notebook_file_path: Location to write notebook
        """
        with open(notebook_file_path, "w") as f:
            nbformat.write(notebook, f)

    def render(self, **kwargs: dict) -> nbformat.NotebookNode:
        """
        Render a notebook from parameters.
        """
        raise NotImplementedError

    def render_to_disk(self, notebook_file_path: str, **kwargs: dict) -> None:
        """
        Render a notebook to disk from arguments.
        """
        raise NotImplementedError



class ProfilerTypeMapping:
    """Useful backend type mapping for building profilers."""

    INT_TYPE_NAMES = [
        "BIGINT",
        "BYTEINT",
        "ByteType()",
        "INT",
        "INT64",
        "INTEGER",
        "Int16Dtype",
        "Int32Dtype",
        "Int64Dtype",
        "Int8Dtype",
        "IntegerType",
        "IntegerType()",
        "LongType",
        "LongType()",
        "SMALLINT",
        "ShortType()",
        "TINYINT",
        "UInt16Dtype",
        "UInt32Dtype",
        "UInt64Dtype",
        "UInt8Dtype",
        "int",
        "int16",
        "int32",
        "int64",
        "int8",
        "int_",
        "integer",
        "uint16",
        "uint32",
        "uint64",
        "uint8",
    ]
    FLOAT_TYPE_NAMES = [
        "DECIMAL",
        "DOUBLE",
        "DOUBLE_PRECISION",
        "DecimalType()",
        "DoubleType",
        "DoubleType()",
        "FLOAT",
        "FLOAT4",
        "FLOAT64",
        "FLOAT8",
        "FloatType",
        "FloatType()",
        "NUMERIC",
        "REAL",
        "float",
        "float16",
        "float32",
        "float64",
        "float_",
        "number",
    ]
    STRING_TYPE_NAMES = [
        "CHAR",
        "NCHAR",
        "NTEXT",
        "NVARCHAR",
        "STRING",
        "StringType",
        "StringType()",
        "TEXT",
        "VARCHAR",
        "dtype('O')",
        "object",
        "str",
        "string",
    ]
    BOOLEAN_TYPE_NAMES = [
        "BIT",
        "BOOL",
        "BOOLEAN",
        "BooleanType",
        "BooleanType()",
        "TINYINT",
        "bool",
        "boolean",
    ]
    DATETIME_TYPE_NAMES = [
        "DATE",
        "TIME",
        "DATETIME",
        "DATETIME2",
        "DATETIME64",
        "SMALLDATETIME",
        "DATETIMEOFFSET",
        "TIMESTAMP",
        "Timestamp",
        "TimestampType",
        "TimestampType()",
        "DateType",
        "DateType()",
        "datetime64",
        "datetime64[ns]",
        "timedelta[ns]",
        "<M8[ns]",
    ]
    BINARY_TYPE_NAMES = [
        "BINARY",
        "BinaryType()",
        "IMAGE",
        "VARBINARY",
        "binary",
        "image",
        "varbinary",
    ]
    CURRENCY_TYPE_NAMES = [
        "MONEY",
        "SMALLMONEY",
        "money",
        "smallmoney",
    ]
    IDENTIFIER_TYPE_NAMES = [
        "UNIQUEIDENTIFIER",
        "uniqueidentifier",
    ]
    MISCELLANEOUS_TYPE_NAMES = [
        "SQL_VARIANT",
        "sql_variant",
    ]
    RECORD_TYPE_NAMES = [
        "JSON",
        "json",
    ]
    OBJECT_TYPE_NAMES = [
        "OBJECT",
        "object",
    ]



from typing import TYPE_CHECKING, Dict, List, Optional

from great_expectations.core import (
    ExpectationConfiguration,  # noqa: TCH001
    ExpectationValidationResult,  # noqa: TCH001
)
from great_expectations.core._docs_decorators import public_api
from great_expectations.exceptions import InvalidExpectationConfigurationError
from great_expectations.execution_engine import ExecutionEngine  # noqa: TCH001
from great_expectations.expectations.expectation import (
    TableExpectation,
    render_evaluation_parameter_string,
)
from great_expectations.render import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import substitute_none_for_missing
from great_expectations.rule_based_profiler.config import (
    ParameterBuilderConfig,
    RuleBasedProfilerConfig,
)
from great_expectations.rule_based_profiler.parameter_container import (
    DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
    FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY,
    FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER,
    FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY,
    PARAMETER_KEY,
    VARIABLES_KEY,
)

if TYPE_CHECKING:
    from great_expectations.render.renderer_configuration import AddParamArgs


class ExpectTableColumnsToMatchSet(TableExpectation):
    """Expect the columns to match an unordered set.

    expect_table_columns_to_match_set is a \
    [Table Expectation](https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_table_expectations).

    Args:
        column_set (list of str): \
            The column names, in any order.
        exact_match (boolean): \
            Whether the list of columns must exactly match the observed columns.

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        include_config (boolean): \
            If True, then include the expectation config as part of the result object.
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, include_config, catch_exceptions, and meta.
    """

    library_metadata = {
        "maturity": "production",
        "tags": ["core expectation", "table expectation"],
        "contributors": [
            "@great_expectations",
        ],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }

    metric_dependencies = ("table.columns",)
    success_keys = (
        "column_set",
        "exact_match",
        "auto",
        "profiler_config",
    )

    mean_table_columns_set_match_multi_batch_parameter_builder_config = (
        ParameterBuilderConfig(
            module_name="great_expectations.rule_based_profiler.parameter_builder",
            class_name="MeanTableColumnsSetMatchMultiBatchParameterBuilder",
            name="column_names_set_estimator",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=None,
        )
    )
    validation_parameter_builder_configs: List[ParameterBuilderConfig] = [
        mean_table_columns_set_match_multi_batch_parameter_builder_config,
    ]
    default_profiler_config = RuleBasedProfilerConfig(
        name="expect_table_columns_to_match_set",  # Convention: use "expectation_type" as profiler name.
        config_version=1.0,
        variables={},
        rules={
            "expect_table_columns_to_match_set": {
                "variables": {
                    "exact_match": None,
                    "success_ratio": 1.0,
                },
                "domain_builder": {
                    "class_name": "TableDomainBuilder",
                    "module_name": "great_expectations.rule_based_profiler.domain_builder",
                },
                "expectation_configuration_builders": [
                    {
                        "expectation_type": "expect_table_columns_to_match_set",
                        "class_name": "DefaultExpectationConfigurationBuilder",
                        "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                        "validation_parameter_builder_configs": validation_parameter_builder_configs,
                        "condition": f"{PARAMETER_KEY}{mean_table_columns_set_match_multi_batch_parameter_builder_config.name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}success_ratio >= {VARIABLES_KEY}success_ratio",
                        "column_set": f"{PARAMETER_KEY}{mean_table_columns_set_match_multi_batch_parameter_builder_config.name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY}",
                        "exact_match": f"{VARIABLES_KEY}exact_match",
                        "meta": {
                            "profiler_details": f"{PARAMETER_KEY}{mean_table_columns_set_match_multi_batch_parameter_builder_config.name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY}",
                        },
                    },
                ],
            },
        },
    )

    default_kwarg_values = {
        "column_set": None,
        "exact_match": True,
        "result_format": "BASIC",
        "include_config": True,
        "catch_exceptions": False,
        "auto": False,
        "profiler_config": default_profiler_config,
    }
    args_keys = (
        "column_set",
        "exact_match",
    )

    @public_api
    def validate_configuration(
        self, configuration: Optional[ExpectationConfiguration] = None
    ) -> None:
        """Validates the configuration of an Expectation.

        For this expectation it is required that:

        - `column_set` has been provided.
        - `column_set` is one of the following types: `list`, `set`, or `None`
        - If `column_set` is a `dict`, it is assumed to be an Evaluation Parameter, and therefore the
          dictionary keys must be `$PARAMETER`.

        The configuration will also be validated using each of the `validate_configuration` methods in its Expectation
        superclass hierarchy.

        Args:
            configuration: An `ExpectationConfiguration` to validate. If no configuration is provided, it will be pulled
                from the configuration attribute of the Expectation instance.

        Raises:
            InvalidExpectationConfigurationError: The configuration does not contain the values required by the
                Expectation.
        """
        # Setting up a configuration
        super().validate_configuration(configuration)

        # Ensuring that a proper value has been provided
        try:
            assert "column_set" in configuration.kwargs, "column_set is required"
            assert (
                isinstance(configuration.kwargs["column_set"], (list, set, dict))
                or configuration.kwargs["column_set"] is None
            ), "column_set must be a list, set, or None"
            if isinstance(configuration.kwargs["column_set"], dict):
                assert (
                    "$PARAMETER" in configuration.kwargs["column_set"]
                ), 'Evaluation Parameter dict for column_set kwarg must have "$PARAMETER" key.'
        except AssertionError as e:
            raise InvalidExpectationConfigurationError(str(e))

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column_set", RendererValueType.ARRAY),
            ("exact_match", RendererValueType.BOOLEAN),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if not params.column_set:
            template_str = "Must specify a set or list of columns."
        else:
            array_param_name = "column_set"
            param_prefix = "column_set_"
            renderer_configuration = cls._add_array_params(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            column_set_str: str = cls._get_array_string(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )

            exact_match_str = (
                "exactly"
                if params.exact_match and params.exact_match.value is True
                else "at least"
            )

            template_str = f"Must have {exact_match_str} these columns (in any order): {column_set_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_evaluation_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        _ = False if runtime_configuration.get("include_column_name") is False else True
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs, ["column_set", "exact_match"]
        )

        if params["column_set"] is None:
            template_str = "Must specify a set or list of columns."

        else:
            # standardize order of the set for output
            params["column_list"] = list(params["column_set"])

            column_list_template_str = ", ".join(
                [f"$column_list_{idx}" for idx in range(len(params["column_list"]))]
            )

            exact_match_str = "exactly" if params["exact_match"] is True else "at least"

            template_str = f"Must have {exact_match_str} these columns (in any order): {column_list_template_str}"

            for idx in range(len(params["column_list"])):
                params[f"column_list_{str(idx)}"] = params["column_list"][idx]

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]

    def _validate(
        self,
        configuration: ExpectationConfiguration,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        # Obtaining columns and ordered list for sake of comparison
        expected_column_set = self.get_success_kwargs(configuration).get("column_set")
        expected_column_set = (
            set(expected_column_set) if expected_column_set is not None else set()
        )
        actual_column_list = metrics.get("table.columns")
        actual_column_set = set(actual_column_list)
        exact_match = self.get_success_kwargs(configuration).get("exact_match")

        if (
            (expected_column_set is None) and (exact_match is not True)
        ) or actual_column_set == expected_column_set:
            return {"success": True, "result": {"observed_value": actual_column_list}}
        else:
            # Convert to lists and sort to lock order for testing and output rendering
            # unexpected_list contains items from the dataset columns that are not in expected_column_set
            unexpected_list = sorted(list(actual_column_set - expected_column_set))
            # missing_list contains items from expected_column_set that are not in the dataset columns
            missing_list = sorted(list(expected_column_set - actual_column_set))
            # observed_value contains items that are in the dataset columns
            observed_value = sorted(actual_column_list)

            mismatched = {}
            if len(unexpected_list) > 0:
                mismatched["unexpected"] = unexpected_list
            if len(missing_list) > 0:
                mismatched["missing"] = missing_list

            result = {
                "observed_value": observed_value,
                "details": {"mismatched": mismatched},
            }

            return_success = {
                "success": True,
                "result": result,
            }
            return_failed = {
                "success": False,
                "result": result,
            }

            if exact_match:
                return return_failed
            else:
                # Failed if there are items in the missing list (but OK to have unexpected_list)
                if len(missing_list) > 0:
                    return return_failed
                # Passed if there are no items in the missing list
                else:
                    return return_success



from typing import List

import pytest

from great_expectations.core.usage_statistics.package_dependencies import GXDependencies


def test__get_dependency_names():
    """Test that the regex in _get_dependency_names() parses a requirements file correctly."""
    mock_dependencies: List[str] = [
        "# This is a comment",
        "#This is another comment",
        "--requirement requirements.txt",
        "latest-package",
        "duplicate-package",
        "duplicate-package",  # Should not be filtered in this method
        "pinned-package1==0.2.4",
        "pinned-package2==0.2.4 # With comment after",
        "lower-bound-package1>=5.2.4",
        "lower-bound-package2>=5.2.4 # With comment after",
        "lower-bound-package3>5.2.4",
        "upper-bound-package1<=2.3.8",
        "upper-bound-package2<=2.3.8 # With comment after",
        "upper-bound-package3<2.3.8",
        "two-bounds-package1>=0.8.4,<2.0.0",
        "two-bounds-package2>=0.8.4,<2.0.0 # With comment after",
        "package_with_underscores",
        "1",
        "-",
        "",
    ]
    expected_dependendencies: List[str] = [
        "latest-package",
        "duplicate-package",
        "duplicate-package",
        "pinned-package1",
        "pinned-package2",
        "lower-bound-package1",
        "lower-bound-package2",
        "lower-bound-package3",
        "upper-bound-package1",
        "upper-bound-package2",
        "upper-bound-package3",
        "two-bounds-package1",
        "two-bounds-package2",
        "package_with_underscores",
        "1",
        "-",
    ]
    ge_dependencies = GXDependencies()
    observed_dependencies = ge_dependencies._get_dependency_names(mock_dependencies)
    assert observed_dependencies == expected_dependendencies


@pytest.mark.integration
def test_required_dependency_names_match_requirements_file():
    """If there is a mismatch, GXDependencies should change to match our requirements.txt file.

    See GXDependencies for a utility to check for a mismatch.
    """
    ge_dependencies = GXDependencies()
    assert (
        ge_dependencies.get_required_dependency_names()
        == ge_dependencies.get_required_dependency_names_from_requirements_file()
    )


@pytest.mark.integration
def test_dev_dependency_names_match_requirements_file():
    """If there is a mismatch, GXDependencies should change to match our requirements-dev*.txt files.

    See GXDependencies for a utility to check for a mismatch.
    """
    ge_dependencies = GXDependencies()
    assert ge_dependencies.get_dev_dependency_names() == set(
        ge_dependencies.get_dev_dependency_names_from_requirements_file()
    ) - set(GXDependencies.GX_DEV_DEPENDENCIES_EXCLUDED_FROM_TRACKING)



import pathlib

import pytest

from great_expectations.compatibility.pydantic import ValidationError
from great_expectations.experimental.datasources import SqliteDatasource
from great_expectations.experimental.datasources.sqlite_datasource import SqliteDsn


@pytest.fixture
def sqlite_datasource_name() -> str:
    return "sqlite_datasource"


@pytest.fixture
def sqlite_database_path() -> pathlib.Path:
    relative_path = pathlib.Path(
        "..",
        "..",
        "test_sets",
        "taxi_yellow_tripdata_samples",
        "sqlite",
        "yellow_tripdata.db",
    )
    return pathlib.Path(__file__).parent.joinpath(relative_path).resolve(strict=True)


@pytest.fixture
def sqlite_datasource(sqlite_database_path, sqlite_datasource_name) -> SqliteDatasource:
    connection_string = f"sqlite:///{sqlite_database_path}"
    datasource = SqliteDatasource(
        name=sqlite_datasource_name,
        connection_string=connection_string,  # type: ignore[arg-type]  # pydantic will coerce
    )
    return datasource


@pytest.mark.unit
def test_connection_string_starts_with_sqlite(
    sqlite_datasource, sqlite_database_path, sqlite_datasource_name
):
    # The actual file doesn't matter only it's existence since SqlAlchemy does a check
    # when it creates the database engine.
    assert sqlite_datasource.name == sqlite_datasource_name
    assert sqlite_datasource.connection_string == f"sqlite:///{sqlite_database_path}"


@pytest.mark.unit
def test_connection_string_that_does_not_start_with_sqlite():
    name = "sqlite_datasource"
    connection_string = "stuff+sqlite:///path/to/database/file.db"
    with pytest.raises(ValidationError) as e:
        SqliteDatasource(
            name=name,
            connection_string=connection_string,
        )
    assert str(e.value) == (
        "1 validation error for SqliteDatasource\n"
        "connection_string\n"
        "  URL scheme not permitted (type=value_error.url.scheme; "
        f"allowed_schemes={SqliteDsn.allowed_schemes})"
    )


@pytest.mark.unit
def test_non_select_query_asset(sqlite_datasource):
    with pytest.raises(ValueError):
        sqlite_datasource.add_query_asset(name="query_asset", query="* from table")



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    List,
    NamedTuple,
    Type,
    Union,
)

from great_expectations.experimental.datasources.signatures import _merge_signatures
from great_expectations.experimental.datasources.type_lookup import TypeLookup

if TYPE_CHECKING:
    from great_expectations.compatibility import pydantic
    from great_expectations.data_context import AbstractDataContext as GXDataContext
    from great_expectations.datasource import BaseDatasource, LegacyDatasource
    from great_expectations.experimental.context import DataContext
    from great_expectations.experimental.datasources import PandasDatasource
    from great_expectations.experimental.datasources.interfaces import (
        DataAsset,
        Datasource,
    )

SourceFactoryFn = Callable[..., "Datasource"]

logger = logging.getLogger(__name__)

DEFAULT_PANDAS_DATASOURCE_NAMES: tuple[str, str] = (
    "default_pandas_datasource",
    "default_pandas_fluent_datasource",
)


class DefaultPandasDatasourceError(Exception):
    pass


class TypeRegistrationError(TypeError):
    pass


class _FieldDetails(NamedTuple):
    default_value: Any
    type_annotation: Type


def _get_field_details(
    model: Type[pydantic.BaseModel], field_name: str
) -> _FieldDetails:
    """Get the default value of the requested field and its type annotation."""
    return _FieldDetails(
        default_value=model.__fields__[field_name].default,
        type_annotation=model.__fields__[field_name].type_,
    )


class _SourceFactories:
    """
    Contains a collection of datasource factory methods in the format `.add_<TYPE_NAME>()`

    Contains a `.type_lookup` dict-like two way mapping between previously registered `Datasource`
    or `DataAsset` types and a simplified name for those types.
    """

    # TODO (kilo59): split DataAsset & Datasource lookups
    type_lookup: ClassVar = TypeLookup()
    __source_factories: ClassVar[Dict[str, SourceFactoryFn]] = {}

    _data_context: Union[DataContext, GXDataContext]

    def __init__(self, data_context: Union[DataContext, GXDataContext]):
        self._data_context = data_context

    @classmethod
    def register_types_and_ds_factory(
        cls,
        ds_type: Type[Datasource],
        factory_fn: SourceFactoryFn,
    ) -> None:
        """
        Add/Register a datasource factory function and all related `Datasource`,
        `DataAsset` and `ExecutionEngine` types.

        Creates mapping table between the `DataSource`/`DataAsset` classes and their
        declared `type` string.


        Example
        -------

        An `.add_pandas_filesystem()` pandas_filesystem factory method will be added to `context.sources`.

        >>> class PandasFilesystemDatasource(_PandasDatasource):
        >>>     type: str = 'pandas_filesystem'
        >>>     asset_types = [FileAsset]
        >>>     execution_engine: PandasExecutionEngine
        """

        # TODO: check that the name is a valid python identifier (and maybe that it is snake_case?)
        ds_type_name = _get_field_details(ds_type, "type").default_value
        if not ds_type_name:
            raise TypeRegistrationError(
                f"`{ds_type.__name__}` is missing a `type` attribute with an assigned string value"
            )

        # rollback type registrations if exception occurs
        with cls.type_lookup.transaction() as type_lookup:

            # TODO: We should namespace the asset type to the datasource so different datasources can reuse asset types.
            cls._register_assets(ds_type, asset_type_lookup=type_lookup)

            cls._register_datasource_and_factory_method(
                ds_type,
                factory_fn=factory_fn,
                ds_type_name=ds_type_name,
                datasource_type_lookup=type_lookup,
            )

    @classmethod
    def _register_datasource_and_factory_method(
        cls,
        ds_type: Type[Datasource],
        factory_fn: SourceFactoryFn,
        ds_type_name: str,
        datasource_type_lookup: TypeLookup,
    ) -> str:
        """
        Register the `Datasource` class and add a factory method for the class on `sources`.
        The method name is pulled from the `Datasource.type` attribute.
        """
        method_name = f"add_{ds_type_name}"
        logger.debug(
            f"2a. Registering {ds_type.__name__} as {ds_type_name} with {method_name}() factory"
        )

        pre_existing = cls.__source_factories.get(method_name)
        if pre_existing:
            raise TypeRegistrationError(
                f"'{ds_type_name}' - `sources.{method_name}()` factory already exists",
            )

        datasource_type_lookup[ds_type] = ds_type_name
        logger.debug(f"'{ds_type_name}' added to `type_lookup`")
        cls.__source_factories[method_name] = factory_fn
        return ds_type_name

    @classmethod
    def _register_assets(cls, ds_type: Type[Datasource], asset_type_lookup: TypeLookup):

        asset_types: List[Type[DataAsset]] = ds_type.asset_types

        if not asset_types:
            logger.warning(
                f"No `{ds_type.__name__}.asset_types` have be declared for the `Datasource`"
            )

        for t in asset_types:
            if t.__name__.startswith("_"):
                logger.debug(
                    f"{t} is private, assuming not intended as a public concrete type. Skipping registration"
                )
                continue
            try:
                asset_type_name = _get_field_details(t, "type").default_value
                if asset_type_name is None:
                    raise TypeError(
                        f"{t.__name__} `type` field must be assigned and cannot be `None`"
                    )
                logger.debug(
                    f"2b. Registering `DataAsset` `{t.__name__}` as {asset_type_name}"
                )
                asset_type_lookup[t] = asset_type_name
            except (AttributeError, KeyError, TypeError) as bad_field_exc:
                raise TypeRegistrationError(
                    f"No `type` field found for `{ds_type.__name__}.asset_types` -> `{t.__name__}` unable to register asset type",
                ) from bad_field_exc

            cls._bind_asset_factory_method_if_not_present(ds_type, t, asset_type_name)

    @classmethod
    def _bind_asset_factory_method_if_not_present(
        cls,
        ds_type: Type[Datasource],
        asset_type: Type[DataAsset],
        asset_type_name: str,
    ):
        add_asset_factory_method_name = f"add_{asset_type_name}_asset"
        asset_factory_defined: bool = add_asset_factory_method_name in ds_type.__dict__

        if not asset_factory_defined:
            logger.debug(
                f"No `{add_asset_factory_method_name}()` method found for `{ds_type.__name__}` generating the method..."
            )

            def _add_asset_factory(
                self: Datasource, name: str, **kwargs
            ) -> pydantic.BaseModel:
                asset = asset_type(name=name, **kwargs)
                return self.add_asset(asset)

            # attr-defined issue
            # https://github.com/python/mypy/issues/12472
            _add_asset_factory.__signature__ = _merge_signatures(  # type: ignore[attr-defined]
                _add_asset_factory, asset_type, exclude={"type"}
            )
            setattr(ds_type, add_asset_factory_method_name, _add_asset_factory)

            def _read_asset_factory(self: Datasource, **kwargs) -> pydantic.BaseModel:
                existing_asset_names: Generator[str, None, None] = (
                    asset_name for asset_name in self.assets.keys()
                )
                asset_name_prefix = f"{asset_type_name}_asset_"
                max_asset_number = 0
                for asset_name in existing_asset_names:
                    try:
                        # this can fail for an asset named like csv_asset_foo
                        asset_number = int(asset_name.split("_")[-1])
                    except ValueError:
                        asset_number = 1
                    if (
                        asset_name.startswith(asset_name_prefix)
                        and asset_number > max_asset_number
                    ):
                        max_asset_number = asset_number

                name = asset_name_prefix + str(max_asset_number + 1)
                asset = asset_type(name=name, **kwargs)
                return self.add_asset(asset)

            _read_asset_factory.__signature__ = _merge_signatures(  # type: ignore[attr-defined]
                _read_asset_factory, asset_type, exclude={"type"}
            )
            read_asset_factory_method_name = f"read_{asset_type_name}"
            setattr(ds_type, read_asset_factory_method_name, _read_asset_factory)

        else:
            logger.debug(
                f"`{add_asset_factory_method_name}()` already defined `{ds_type.__name__}`"
            )

    @property
    def pandas_default(self) -> PandasDatasource:
        from great_expectations.experimental.datasources import PandasDatasource

        datasources: dict[
            str, LegacyDatasource | BaseDatasource | Datasource
        ] = self._data_context.datasources  # type: ignore[union-attr]  # typing information is being lost in DataContext factory

        existing_datasource: LegacyDatasource | BaseDatasource | Datasource | None = (
            None
        )
        # datasource names will be attempted in the order they are listed
        for default_pandas_datasource_name in DEFAULT_PANDAS_DATASOURCE_NAMES:
            # if a legacy datasource with this name already exists, we try a different name
            existing_datasource = datasources.get(default_pandas_datasource_name)
            if not existing_datasource or isinstance(
                existing_datasource, PandasDatasource
            ):
                break

        # if a legacy datasource exists for all possible_default_datasource_names, raise an error
        if existing_datasource and not isinstance(
            existing_datasource, PandasDatasource
        ):
            quoted_datasource_names = [
                f'"{name}"' for name in DEFAULT_PANDAS_DATASOURCE_NAMES
            ]
            raise DefaultPandasDatasourceError(
                f"Datasources with a legacy type already exist with the names: {', '.join(quoted_datasource_names)}. "
                "Please rename these datasources if you wish to use the pandas_default PandasDatasource."
            )

        pandas_datasource = (
            existing_datasource
            or self._data_context.sources.add_pandas(
                name=default_pandas_datasource_name
            )
        )
        # there is no situation in which this isn't true
        # return type information must be missing for factory method add_pandas
        assert isinstance(pandas_datasource, PandasDatasource)
        return pandas_datasource

    @property
    def factories(self) -> List[str]:
        return list(self.__source_factories.keys())

    def __getattr__(self, attr_name: str):
        try:
            ds_constructor = self.__source_factories[attr_name]

            def wrapped(name: str, **kwargs):
                datasource = ds_constructor(name=name, **kwargs)
                # TODO (bdirks): _attach_datasource_to_context to the AbstractDataContext class
                self._data_context._attach_datasource_to_context(datasource)
                return datasource

            wrapped.__doc__ = ds_constructor.__doc__
            return wrapped
        except KeyError:
            raise AttributeError(f"No factory {attr_name} in {self.factories}")

    def __dir__(self) -> List[str]:
        """Preserves autocompletion for dynamic attributes."""
        return [*self.factories, *super().__dir__()]



from .data_connector import DataConnector
from .file_path_data_connector import FilePathDataConnector
from .filesystem_data_connector import FilesystemDataConnector
from .s3_data_connector import S3DataConnector



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    ClassVar,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.domain import Domain, SemanticDomainTypes
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.rule_based_profiler.domain_builder import DomainBuilder
from great_expectations.rule_based_profiler.helpers.util import (
    build_domains_from_column_names,
    get_parameter_value_and_validate_return_type,
)
from great_expectations.rule_based_profiler.parameter_container import (
    ParameterContainer,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.semantic_type_filter import (
    SemanticTypeFilter,  # noqa: TCH001
)
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )
    from great_expectations.validator.validator import Validator


class ColumnDomainBuilder(DomainBuilder):
    exclude_field_names: ClassVar[Set[str]] = DomainBuilder.exclude_field_names | {
        "semantic_type_filter",
    }

    def __init__(
        self,
        include_column_names: Optional[Union[str, Optional[List[str]]]] = None,
        exclude_column_names: Optional[Union[str, Optional[List[str]]]] = None,
        include_column_name_suffixes: Optional[Union[str, Iterable, List[str]]] = None,
        exclude_column_name_suffixes: Optional[Union[str, Iterable, List[str]]] = None,
        semantic_type_filter_module_name: Optional[str] = None,
        semantic_type_filter_class_name: Optional[str] = None,
        include_semantic_types: Optional[
            Union[str, SemanticDomainTypes, List[Union[str, SemanticDomainTypes]]]
        ] = None,
        exclude_semantic_types: Optional[
            Union[str, SemanticDomainTypes, List[Union[str, SemanticDomainTypes]]]
        ] = None,
        data_context: Optional[AbstractDataContext] = None,
    ) -> None:
        """
        A semantic type is distinguished from the structured column type;
        An example structured column type would be "integer".  The inferred semantic type would be "id".

        Args:
            include_column_names: Explicitly specified desired columns (if None, it is computed based on active Batch).
            exclude_column_names: If provided, these columns are pre-filtered and excluded from consideration.
            include_column_name_suffixes: Explicitly specified desired suffixes for corresponding columns to match.
            exclude_column_name_suffixes: Explicitly specified desired suffixes for corresponding columns to not match.
            semantic_type_filter_module_name: module_name containing class that implements SemanticTypeFilter interfaces
            semantic_type_filter_class_name: class_name of class that implements SemanticTypeFilter interfaces
            include_semantic_types: single/multiple type specifications using SemanticDomainTypes (or str equivalents)
            to be included
            exclude_semantic_types: single/multiple type specifications using SemanticDomainTypes (or str equivalents)
            to be excluded
            data_context: AbstractDataContext associated with this DomainBuilder

        Inclusion/Exclusion Logic:
        (include_column_names|table_columns - exclude_column_names) + (include_semantic_types - exclude_semantic_types)
        """
        super().__init__(data_context=data_context)

        self._include_column_names = include_column_names
        self._exclude_column_names = exclude_column_names

        self._include_column_name_suffixes = include_column_name_suffixes
        self._exclude_column_name_suffixes = exclude_column_name_suffixes

        self._semantic_type_filter_module_name = semantic_type_filter_module_name
        self._semantic_type_filter_class_name = semantic_type_filter_class_name

        self._include_semantic_types = include_semantic_types
        self._exclude_semantic_types = exclude_semantic_types

        self._semantic_type_filter = None

    @property
    def domain_type(self) -> MetricDomainTypes:
        return MetricDomainTypes.COLUMN

    """
    All DomainBuilder classes, whose "domain_type" property equals "MetricDomainTypes.COLUMN", must extend present class
    (ColumnDomainBuilder) in order to provide full getter/setter accessor for relevant properties (as overrides).
    """

    @property
    def include_column_names(self) -> Optional[Union[str, Optional[List[str]]]]:
        return self._include_column_names

    @include_column_names.setter
    def include_column_names(
        self, value: Optional[Union[str, Optional[List[str]]]]
    ) -> None:
        self._include_column_names = value

    @property
    def exclude_column_names(self) -> Optional[Union[str, Optional[List[str]]]]:
        return self._exclude_column_names

    @exclude_column_names.setter
    def exclude_column_names(
        self, value: Optional[Union[str, Optional[List[str]]]]
    ) -> None:
        self._exclude_column_names = value

    @property
    def include_column_name_suffixes(
        self,
    ) -> Optional[Union[str, Iterable, List[str]]]:
        return self._include_column_name_suffixes

    @include_column_name_suffixes.setter
    def include_column_name_suffixes(
        self, value: Optional[Union[str, Iterable, List[str]]]
    ) -> None:
        self._include_column_name_suffixes = value

    @property
    def exclude_column_name_suffixes(
        self,
    ) -> Optional[Union[str, Iterable, List[str]]]:
        return self._exclude_column_name_suffixes

    @exclude_column_name_suffixes.setter
    def exclude_column_name_suffixes(
        self, value: Optional[Union[str, Iterable, List[str]]]
    ) -> None:
        self._exclude_column_name_suffixes = value

    @property
    def semantic_type_filter_module_name(self) -> Optional[str]:
        return self._semantic_type_filter_module_name

    @property
    def semantic_type_filter_class_name(self) -> Optional[str]:
        return self._semantic_type_filter_class_name

    @property
    def include_semantic_types(
        self,
    ) -> Optional[
        Union[str, SemanticDomainTypes, List[Union[str, SemanticDomainTypes]]]
    ]:
        return self._include_semantic_types

    @include_semantic_types.setter
    def include_semantic_types(
        self,
        value: Optional[
            Union[str, SemanticDomainTypes, List[Union[str, SemanticDomainTypes]]]
        ],
    ) -> None:
        self._include_semantic_types = value

    @property
    def exclude_semantic_types(
        self,
    ) -> Optional[
        Union[str, SemanticDomainTypes, List[Union[str, SemanticDomainTypes]]]
    ]:
        return self._exclude_semantic_types

    @exclude_semantic_types.setter
    def exclude_semantic_types(
        self,
        value: Optional[
            Union[str, SemanticDomainTypes, List[Union[str, SemanticDomainTypes]]]
        ],
    ) -> None:
        self._exclude_semantic_types = value

    @property
    def semantic_type_filter(self) -> Optional[SemanticTypeFilter]:
        return self._semantic_type_filter

    def get_effective_column_names(
        self,
        batch_ids: Optional[List[str]] = None,
        validator: Optional[Validator] = None,
        variables: Optional[ParameterContainer] = None,
    ) -> List[str]:
        """
        This method applies multiple directives to obtain columns to be included as part of returned "Domain" objects.
        """
        include_column_names: List[str] = cast(
            List[str],
            self._resolve_list_type_property(
                property_name="include_column_names",
                property_value_type=list,
                variables=variables,
            ),
        )

        if batch_ids is None:
            batch_ids: List[str] = self.get_batch_ids(variables=variables)  # type: ignore[no-redef]

        if validator is None:
            validator = self.get_validator(variables=variables)

        table_columns: List[str] = validator.get_metric(  # type: ignore[union-attr] # could be None
            metric=MetricConfiguration(
                metric_name="table.columns",
                metric_domain_kwargs={
                    # active_batch_id
                    "batch_id": batch_ids[-1],  # type: ignore[index]
                },
                metric_value_kwargs=None,
            )
        )

        effective_column_names: List[str] = include_column_names or table_columns

        exclude_column_names: List[str] = cast(
            List[str],
            self._resolve_list_type_property(
                property_name="exclude_column_names",
                property_value_type=list,
                variables=variables,
            ),
        )

        column_name: str

        effective_column_names = [
            column_name
            for column_name in effective_column_names
            if column_name not in exclude_column_names
        ]

        for column_name in effective_column_names:
            if column_name not in table_columns:
                raise gx_exceptions.ProfilerExecutionError(
                    message=f'Error: The column "{column_name}" in BatchData does not exist.'
                )

        include_column_name_suffixes: List[str] = cast(
            List[str],
            self._resolve_list_type_property(
                property_name="include_column_name_suffixes",
                property_value_type=(str, Iterable, list),
                variables=variables,
            ),
        )
        if include_column_name_suffixes:
            effective_column_names = list(
                filter(
                    lambda candidate_column_name: candidate_column_name.endswith(
                        tuple(include_column_name_suffixes)
                    ),
                    effective_column_names,
                )
            )

        exclude_column_name_suffixes: List[str] = cast(
            List[str],
            self._resolve_list_type_property(
                property_name="exclude_column_name_suffixes",
                property_value_type=(str, Iterable, list),
                variables=variables,
            ),
        )
        if exclude_column_name_suffixes:
            effective_column_names = list(
                filter(
                    lambda candidate_column_name: not candidate_column_name.endswith(
                        tuple(exclude_column_name_suffixes)
                    ),
                    effective_column_names,
                )
            )

        # Obtain semantic_type_filter_module_name from "rule state" (i.e., variables and parameters); from instance variable otherwise.
        semantic_type_filter_module_name: Optional[
            str
        ] = get_parameter_value_and_validate_return_type(
            domain=None,
            parameter_reference=self.semantic_type_filter_module_name,
            expected_return_type=None,
            variables=variables,
            parameters=None,
        )
        if semantic_type_filter_module_name is None:
            semantic_type_filter_module_name = "great_expectations.rule_based_profiler.helpers.simple_semantic_type_filter"

        # Obtain semantic_type_filter_class_name from "rule state" (i.e., variables and parameters); from instance variable otherwise.
        semantic_type_filter_class_name: Optional[
            str
        ] = get_parameter_value_and_validate_return_type(
            domain=None,
            parameter_reference=self.semantic_type_filter_class_name,
            expected_return_type=None,
            variables=variables,
            parameters=None,
        )
        if semantic_type_filter_class_name is None:
            semantic_type_filter_class_name = "SimpleSemanticTypeFilter"

        semantic_type_filter: SemanticTypeFilter = instantiate_class_from_config(
            config={
                "module_name": semantic_type_filter_module_name,
                "class_name": semantic_type_filter_class_name,
            },
            runtime_environment={
                "batch_ids": batch_ids,
                "validator": validator,
                "column_names": effective_column_names,
            },
            config_defaults={},
        )
        self._semantic_type_filter = semantic_type_filter  # type: ignore[assignment] # could be None

        include_semantic_types: List[Union[str, SemanticDomainTypes]] = cast(
            List[Union[str, SemanticDomainTypes]],
            self._resolve_list_type_property(
                property_name="include_semantic_types",
                property_value_type=(str, SemanticDomainTypes, list),
                variables=variables,
            ),
        )
        include_semantic_types = self.semantic_type_filter.parse_semantic_domain_type_argument(  # type: ignore[union-attr,assignment] # could be None
            semantic_types=include_semantic_types
        )

        if include_semantic_types:
            effective_column_names = list(
                filter(
                    lambda candidate_column_name: self.semantic_type_filter.table_column_name_to_inferred_semantic_domain_type_map[  # type: ignore[union-attr,arg-type]
                        candidate_column_name
                    ]
                    in include_semantic_types,
                    effective_column_names,
                )
            )

        exclude_semantic_types: List[Union[str, SemanticDomainTypes]] = cast(
            List[Union[str, SemanticDomainTypes]],
            self._resolve_list_type_property(
                property_name="exclude_semantic_types",
                property_value_type=(str, SemanticDomainTypes, list),
                variables=variables,
            ),
        )
        exclude_semantic_types = self.semantic_type_filter.parse_semantic_domain_type_argument(  # type: ignore[union-attr,assignment] # could be None
            semantic_types=exclude_semantic_types
        )

        if exclude_semantic_types:
            effective_column_names = list(
                filter(
                    lambda candidate_column_name: self.semantic_type_filter.table_column_name_to_inferred_semantic_domain_type_map[  # type: ignore[union-attr,arg-type] # lambda missing type details
                        candidate_column_name
                    ]
                    not in exclude_semantic_types,
                    effective_column_names,
                )
            )

        return effective_column_names

    def _get_domains(
        self,
        rule_name: str,
        variables: Optional[ParameterContainer] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> List[Domain]:
        """
        Obtains and returns domains for all columns of a table (or for configured columns, if they exist in the table).

        Args:
            rule_name: name of Rule object, for which "Domain" objects are obtained.
            variables: Optional variables to substitute when evaluating.
            runtime_configuration: Optional[dict] = None,

        Returns:
            List of domains that match the desired columns and filtering criteria.
        """
        batch_ids: List[str] = self.get_batch_ids(variables=variables)  # type: ignore[assignment] # could be None

        validator: Validator = self.get_validator(variables=variables)  # type: ignore[assignment] # could be None

        effective_column_names: List[str] = self.get_effective_column_names(
            batch_ids=batch_ids,
            validator=validator,
            variables=variables,
        )

        column_name: str
        domains: List[Domain] = build_domains_from_column_names(
            rule_name=rule_name,
            column_names=effective_column_names,
            domain_type=self.domain_type,
            table_column_name_to_inferred_semantic_domain_type_map=self.semantic_type_filter.table_column_name_to_inferred_semantic_domain_type_map,  # type: ignore[union-attr] # could be None
        )

        return domains

    def _resolve_list_type_property(
        self,
        property_name: str,
        property_value_type: Union[type, Tuple[type, ...]],
        variables: Optional[ParameterContainer] = None,
    ) -> List[type]:
        property_value = getattr(self, property_name, [])
        if property_value is None:
            property_value = []
        elif isinstance(property_value, str):
            property_value = [property_value]
        else:
            if not isinstance(property_value, property_value_type):
                raise ValueError(
                    f'Unrecognized "{property_name}" directive -- must be "{property_value_type}" (or string).'
                )

        property_cursor: type
        property_value = [
            # Obtain property from "rule state" (i.e., variables and parameters); from instance variable otherwise.
            get_parameter_value_and_validate_return_type(
                domain=None,
                parameter_reference=property_cursor,
                expected_return_type=None,
                variables=variables,
                parameters=None,
            )
            for property_cursor in property_value
        ]

        return property_value



import pathlib
import re
from typing import List

import pytest

from great_expectations.core import IDDict
from great_expectations.core.batch import BatchDefinition
from great_expectations.experimental.datasources.data_asset.data_connector import (
    DataConnector,
    FilesystemDataConnector,
)
from great_expectations.experimental.datasources.interfaces import BatchRequest
from tests.test_utils import create_files_in_directory


@pytest.mark.integration
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_basic_instantiation(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_basic_instantiation"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "alpha-1.csv",
            "alpha-2.csv",
            "alpha-3.csv",
        ],
    )

    my_data_connector: DataConnector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
        batching_regex=re.compile(r"alpha-(.*)\.csv"),
        glob_directive="*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    # Missing "data_asset_name" argument.
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list(
            BatchRequest(
                datasource_name="something",
                options={},
            )
        )


@pytest.mark.integration
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_instantiation_batching_regex_does_not_match_paths(tmp_path_factory):
    base_directory = str(
        tmp_path_factory.mktemp(
            "test_instantiation_from_a_config_batching_regex_does_not_match_paths"
        )
    )
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "alpha-1.csv",
            "alpha-2.csv",
            "alpha-3.csv",
        ],
    )

    my_data_connector: DataConnector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
        batching_regex=re.compile(r"beta-(.*)\.csv"),
        glob_directive="*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert len(my_data_connector.get_unmatched_data_references()) == 3


@pytest.mark.integration
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_return_all_batch_definitions_unsorted(tmp_path_factory):
    base_directory = str(
        tmp_path_factory.mktemp("test_return_all_batch_definitions_unsorted")
    )
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "alex_20200809_1000.csv",
            "eugene_20200809_1500.csv",
            "james_20200811_1009.csv",
            "abe_20200809_1040.csv",
            "will_20200809_1002.csv",
            "james_20200713_1567.csv",
            "eugene_20201129_1900.csv",
            "will_20200810_1001.csv",
            "james_20200810_1003.csv",
            "alex_20200819_1300.csv",
        ],
    )

    my_data_connector: DataConnector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
        batching_regex=re.compile(r"(?P<name>.+)_(?P<timestamp>.+)_(?P<price>.+)\.csv"),
        glob_directive="*.csv",
    )
    # with missing BatchRequest arguments
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list()

    # with empty options
    unsorted_batch_definition_list: List[
        BatchDefinition
    ] = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_dataframe_datasource",
            data_asset_name="my_filesystem_data_asset",
            options={},
        )
    )
    expected: List[BatchDefinition] = [
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "abe", "timestamp": "20200809", "price": "1040"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "alex", "timestamp": "20200809", "price": "1000"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "alex", "timestamp": "20200819", "price": "1300"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "eugene", "timestamp": "20200809", "price": "1500"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "eugene", "timestamp": "20201129", "price": "1900"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "james", "timestamp": "20200713", "price": "1567"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "james", "timestamp": "20200810", "price": "1003"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "james", "timestamp": "20200811", "price": "1009"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "will", "timestamp": "20200809", "price": "1002"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict(
                {"name": "will", "timestamp": "20200810", "price": "1001"}
            ),
        ),
    ]
    assert expected == unsorted_batch_definition_list

    # with specified Batch query options
    unsorted_batch_definition_list = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_dataframe_datasource",
            data_asset_name="my_filesystem_data_asset",
            options={"name": "alex", "timestamp": "20200819", "price": "1300"},
        )
    )
    assert expected[2:3] == unsorted_batch_definition_list


# TODO: <Alex>ALEX-UNCOMMENT_WHEN_SORTERS_ARE_INCLUDED_AND_TEST_SORTED_BATCH_DEFINITION_LIST</Alex>
# TODO: <Alex>ALEX</Alex>
# @pytest.mark.integration
# @pytest.mark.slow  # creating small number of`file handles in temporary file system
# def test_return_all_batch_definitions_sorted(tmp_path_factory):
#     base_directory = str(
#         tmp_path_factory.mktemp("test_return_all_batch_definitions_sorted")
#     )
#     create_files_in_directory(
#         directory=base_directory,
#         file_name_list=[
#             "alex_20200809_1000.csv",
#             "eugene_20200809_1500.csv",
#             "james_20200811_1009.csv",
#             "abe_20200809_1040.csv",
#             "will_20200809_1002.csv",
#             "james_20200713_1567.csv",
#             "eugene_20201129_1900.csv",
#             "will_20200810_1001.csv",
#             "james_20200810_1003.csv",
#             "alex_20200819_1300.csv",
#         ],
#     )
#
#     my_data_connector: DataConnector = FilesystemDataConnector(
#         datasource_name="my_dataframe_datasource",
#         data_asset_name="my_filesystem_data_asset",
#         base_directory=pathlib.Path(base_directory),
#         batching_regex=re.compile(r"(?P<name>.+)_(?P<timestamp>.+)_(?P<price>.+)\.csv"),
#         glob_directive="*.csv",
#     )
#     assert my_data_connector.get_data_reference_count() == 3
#     assert my_data_connector._get_data_reference_list()[:3] == [
#         "alpha-1.csv",
#         "alpha-2.csv",
#         "alpha-3.csv",
#     ]
#     assert my_data_connector.get_unmatched_data_references()[:3] == [
#         "alpha-1.csv",
#         "alpha-2.csv",
#         "alpha-3.csv",
#     ]
#     assert len(my_data_connector.get_unmatched_data_references()) == 3
#
#     sorted_batch_definition_list: List[BatchDefinition] = (
#         my_data_connector.get_batch_definition_list(
#             BatchRequest(
#                 datasource_name="my_dataframe_datasource",
#                 data_asset_name="my_filesystem_data_asset",
#                 options={},
#             )
#         )
#     )
#
#     expected: List[BatchDefinition] = [
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "abe", "timestamp": "20200809", "price": "1040"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "alex", "timestamp": "20200819", "price": "1300"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "alex", "timestamp": "20200809", "price": "1000"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "eugene", "timestamp": "20201129", "price": "1900"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "eugene", "timestamp": "20200809", "price": "1500"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "james", "timestamp": "20200811", "price": "1009"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "james", "timestamp": "20200810", "price": "1003"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "james", "timestamp": "20200713", "price": "1567"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "will", "timestamp": "20200810", "price": "1001"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_filesystem_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "will", "timestamp": "20200809", "price": "1002"}
#             ),
#         ),
#     ]
#
#     # TEST 1: Sorting works
#     assert expected == sorted_batch_definition_list
#
#     my_batch_request: BatchRequest = BatchRequest(
#         datasource_name="my_dataframe_datasource",
#         data_asset_name="my_filesystem_data_asset",
#         options={
#             "name": "james",
#             "timestamp": "20200713",
#             "price": "1567",
#         },
#     )
#
#     my_batch_definition_list: List[BatchDefinition]
#     my_batch_definition: BatchDefinition
#
#     # TEST 2: Should only return the specified partition
#     my_batch_definition_list = (
#         my_data_connector.get_batch_definition_list(
#             batch_request=my_batch_request
#         )
#     )
#     assert len(my_batch_definition_list) == 1
#     my_batch_definition = my_batch_definition_list[0]
#
#     expected_batch_definition = BatchDefinition(
#         datasource_name="my_dataframe_datasource",
#         data_asset_name="my_filesystem_data_asset",
#         batch_identifiers={
#             "name": "james",
#             "timestamp": "20200713",
#             "price": "1567",
#         },
#     )
#     assert my_batch_definition == expected_batch_definition
#
#     # TEST 3: Without BatchRequest (query) options, should return all 10
#     my_batch_request: BatchRequest = BatchRequest(
#         datasource_name="my_dataframe_datasource",
#         data_asset_name="my_filesystem_data_asset",
#         options={},
#     )
#     # should return 10
#     my_batch_definition_list = (
#         my_data_connector.get_batch_definition_list(
#             batch_request=my_batch_request
#         )
#     )
#     assert len(my_batch_definition_list) == 10
# TODO: <Alex>ALEX</Alex>


@pytest.mark.integration
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_return_only_unique_batch_definitions(tmp_path_factory):
    base_directory = str(
        tmp_path_factory.mktemp("test_return_only_unique_batch_definitions")
    )
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "A/file_1.csv",
            "A/file_2.csv",
            "A/file_3.csv",
            "B/file_1.csv",
            "B/file_2.csv",
        ],
    )

    my_data_connector: DataConnector

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
        batching_regex=re.compile(r"(?P<name>.+)/.+\.csv"),
        # glob_directive="*.csv",  # omitting for purposes of this test
    )
    assert my_data_connector.get_data_reference_count() == 7
    assert my_data_connector.get_data_references()[:3] == [
        "A",
        "A/file_1.csv",
        "A/file_2.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == [
        "A",
        "B",
    ]
    assert len(my_data_connector.get_unmatched_data_references()) == 2

    expected: List[BatchDefinition] = [
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict({"directory": "A", "filename": "file_1.csv"}),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict({"directory": "A", "filename": "file_2.csv"}),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict({"directory": "A", "filename": "file_3.csv"}),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict({"directory": "B", "filename": "file_1.csv"}),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_filesystem_data_asset",
            batch_identifiers=IDDict({"directory": "B", "filename": "file_2.csv"}),
        ),
    ]

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory),
        batching_regex=re.compile(r"(?P<directory>.+)/(?P<filename>.+\.csv)"),
        # glob_directive="*.csv",  # omitting for purposes of this test
    )
    unsorted_batch_definition_list: List[
        BatchDefinition
    ] = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_dataframe_datasource",
            data_asset_name="my_filesystem_data_asset",
            options={},
        )
    )
    assert expected == unsorted_batch_definition_list


@pytest.mark.integration
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_alpha(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_alpha"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_alpha/A.csv",
            "test_dir_alpha/B.csv",
            "test_dir_alpha/C.csv",
            "test_dir_alpha/D.csv",
        ],
    )

    my_data_connector: DataConnector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_alpha",
        batching_regex=re.compile(r"(?P<part_1>.+)\.csv"),
        glob_directive="*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 4
    assert my_data_connector.get_data_references()[:3] == [
        "A.csv",
        "B.csv",
        "C.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_batch_definition_list: List[BatchDefinition]
    my_batch_definition: BatchDefinition

    my_batch_request: BatchRequest

    # Try to fetch a batch from a nonexistent asset
    my_batch_request = BatchRequest(
        datasource_name="BASE", data_asset_name="A", options={}
    )
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 0

    my_batch_request = BatchRequest(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        options={"part_1": "B"},
    )
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 1


@pytest.mark.integration
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_foxtrot(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_foxtrot"))
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_foxtrot/A/A-1.csv",
            "test_dir_foxtrot/A/A-2.csv",
            "test_dir_foxtrot/A/A-3.csv",
            "test_dir_foxtrot/B/B-1.txt",
            "test_dir_foxtrot/B/B-2.txt",
            "test_dir_foxtrot/B/B-3.txt",
            "test_dir_foxtrot/C/C-2017.csv",
            "test_dir_foxtrot/C/C-2018.csv",
            "test_dir_foxtrot/C/C-2019.csv",
            "test_dir_foxtrot/D/D-aaa.csv",
            "test_dir_foxtrot/D/D-bbb.csv",
            "test_dir_foxtrot/D/D-ccc.csv",
            "test_dir_foxtrot/D/D-ddd.csv",
            "test_dir_foxtrot/D/D-eee.csv",
        ],
    )

    my_data_connector: DataConnector

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_foxtrot",
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.csv"),
        glob_directive="*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 0
    assert my_data_connector.get_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_foxtrot" / "A",
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.csv"),
        glob_directive="*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "A-1.csv",
        "A-2.csv",
        "A-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_foxtrot" / "B",
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.txt"),
        glob_directive="*.*",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "B-1.txt",
        "B-2.txt",
        "B-3.txt",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_foxtrot" / "C",
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.csv"),
        glob_directive="*",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "C-2017.csv",
        "C-2018.csv",
        "C-2019.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_batch_request = BatchRequest(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        options={},
    )
    my_batch_definition_list: List[
        BatchDefinition
    ] = my_data_connector.get_batch_definition_list(batch_request=my_batch_request)
    assert len(my_batch_definition_list) == 3


@pytest.mark.integration
@pytest.mark.slow  # creating small number of`file handles in temporary file system
def test_relative_base_directory_path(tmp_path_factory):
    base_directory = str(
        tmp_path_factory.mktemp("test_relative_asset_base_directory_path")
    )
    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_0/A/B/C/logfile_0.csv",
            "test_dir_0/A/B/C/bigfile_1.csv",
            "test_dir_0/A/filename2.csv",
            "test_dir_0/A/filename3.csv",
        ],
    )

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_0" / "A",
        batching_regex=re.compile(r"(?P<part_1>.+)\.csv"),
        glob_directive="*",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "B",
        "filename2.csv",
        "filename3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == [
        "B",
    ]
    assert len(my_data_connector.get_unmatched_data_references()) == 1

    my_data_connector = FilesystemDataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(base_directory) / "test_dir_0" / "A" / "B" / "C",
        batching_regex=re.compile(r"(?P<name>.+)_(?P<number>.+)\.csv"),
        glob_directive="log*.csv",
    )
    assert my_data_connector.get_data_reference_count() == 1
    assert my_data_connector.get_data_references()[:3] == ["logfile_0.csv"]
    assert len(my_data_connector.get_unmatched_data_references()) == 0
    assert (
        my_data_connector._get_full_file_path(path="bigfile_1.csv")
        == f"{base_directory}/test_dir_0/A/B/C/bigfile_1.csv"
    )

    my_batch_request: BatchRequest = BatchRequest(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_filesystem_data_asset",
        options={},
    )
    my_batch_definition_list: List[
        BatchDefinition
    ] = my_data_connector.get_batch_definition_list(batch_request=my_batch_request)
    assert len(my_batch_definition_list) == 1


# TODO: <Alex>ALEX-UNCOMMENT_WHEN_SORTERS_ARE_INCLUDED_AND_TEST_SORTED_BATCH_DEFINITION_LIST</Alex>
# TODO: <Alex>ALEX</Alex>
# def test_return_all_batch_definitions_sorted_sorter_named_that_does_not_match_group(
#     tmp_path_factory,
# ):
#     base_directory = str(
#         tmp_path_factory.mktemp(
#             "test_return_all_batch_definitions_sorted_sorter_named_that_does_not_match_group"
#         )
#     )
#     create_files_in_directory(
#         directory=base_directory,
#         file_name_list=[
#             "alex_20200809_1000.csv",
#             "eugene_20200809_1500.csv",
#             "james_20200811_1009.csv",
#             "abe_20200809_1040.csv",
#             "will_20200809_1002.csv",
#             "james_20200713_1567.csv",
#             "eugene_20201129_1900.csv",
#             "will_20200810_1001.csv",
#             "james_20200810_1003.csv",
#             "alex_20200819_1300.csv",
#         ],
#     )
#     my_data_connector_yaml = yaml.load(
#         f"""
#         class_name: FilesystemDataConnector
#         datasource_name: test_environment
#         base_directory: {base_directory}
#         glob_directive: "*.csv"
#         assets:
#             my_filesystem_data_asset:
#                 pattern: (.+)_(.+)_(.+)\\.csv
#                 group_names:
#                     - name
#                     - timestamp
#                     - price
#         default_regex:
#             pattern: (.+)_.+_.+\\.csv
#             group_names:
#                 - name
#         sorters:
#             - orderby: asc
#               class_name: LexicographicSorter
#               name: name
#             - datetime_format: "%Y%m%d"
#               orderby: desc
#               class_name: DateTimeSorter
#               name: timestamp
#             - orderby: desc
#               class_name: NumericSorter
#               name: for_me_Me_Me
#     """,
#     )
#     with pytest.raises(gx_exceptions.DataConnectorError):
#         # noinspection PyUnusedLocal
#         my_data_connector: FilesystemDataConnector = (
#             instantiate_class_from_config(
#                 config=my_data_connector_yaml,
#                 runtime_environment={
#                     "name": "experimental",
#                     "execution_engine": PandasExecutionEngine(),
#                 },
#                 config_defaults={
#                     "module_name": "great_expectations.datasource.data_connector"
#                 },
#             )
#         )
# TODO: <Alex>ALEX</Alex>



#!/usr/bin/env python3

"""
Test performance using bigquery.
"""

import cProfile
import os
import sys
from collections.abc import Mapping
from pathlib import Path

import _pytest.config
import py.path
import pytest
from pytest_benchmark.fixture import BenchmarkFixture

from great_expectations.checkpoint.types.checkpoint_result import CheckpointResult
from great_expectations.core.async_executor import patch_https_connection_pool
from tests.performance import taxi_benchmark_util

patch_https_connection_pool(taxi_benchmark_util.concurrency_config())


@pytest.mark.parametrize(
    "backend_api",
    [
        "V2",  # Batch Kwargs API
        "V3",  # Batch Request API
    ],
)
@pytest.mark.parametrize("write_data_docs", [False, True])
@pytest.mark.parametrize("number_of_tables", [1, 2, 4, 8, 16, 100])
def test_taxi_trips_benchmark(
    benchmark: BenchmarkFixture,
    tmpdir: py.path.local,
    pytestconfig: _pytest.config.Config,
    number_of_tables: int,
    write_data_docs: bool,
    backend_api: str,
):
    """Benchmark performance with a variety of expectations using NYC Taxi data (yellow_tripdata_sample_2019-01.csv)
    found in the tests/test_sets/taxi_yellow_tripdata_samples directory, and used extensively in unittest and
    integration tests for Great Expectations.

    To simulate a more realistic usage of Great Expectations with several tables, this benchmark is run with 1 or more
    copies of the table, and each table has multiple expectations run on them. For simplicity, the expectations run on
    each table are identical. The specific expectations are somewhat arbitrary but were chosen to be representative of
    a (non-public) real use case of Great Expectations.

    Note: This data being tested in this benchmark generally shouldn't be changed over time, because consistent
    benchmarks are more useful to compare trends over time. Please do not change the tables being tested with nor change
    the expectations being used by this benchmark. Instead of changing this benchmark's data/expectations, please
    consider adding a new benchmark (or at least rename this benchmark to provide clarity that results are not directly
    comparable because of the data change).
    """
    _skip_if_bigquery_performance_tests_not_enabled(pytestconfig)

    html_dir = (
        os.environ.get("GE_BENCHMARK_HTML_DIRECTORY", tmpdir.strpath)
        if write_data_docs
        else None
    )

    checkpoint = taxi_benchmark_util.create_checkpoint(
        number_of_tables=number_of_tables,
        html_dir=html_dir,
        backend_api=backend_api,
    )
    if os.environ.get("GE_PROFILE_FILE_PATH"):
        cProfile.runctx(
            "checkpoint.run()",
            None,
            locals(),
            filename=os.environ["GE_PROFILE_FILE_PATH"],
        )
        return
    else:
        result: CheckpointResult = benchmark.pedantic(
            checkpoint.run,
            iterations=1,
            rounds=1,
        )

    # Do some basic sanity checks.
    assert result.success, result
    assert len(result.run_results) == number_of_tables
    if write_data_docs:
        html_file_paths = list(Path(html_dir).glob("validations/**/*.html"))
        assert len(html_file_paths) == number_of_tables

    # Check that run results contain the right number of suites, assets, and table names.
    assert (
        len(
            {
                run_result["validation_result"]["meta"]["expectation_suite_name"]
                for run_result in result.run_results.values()
            }
        )
        == number_of_tables
    )
    batch_key = "batch_spec" if backend_api == "V3" else "batch_kwargs"
    for field in ["data_asset_name", "table_name" if backend_api == "V3" else "table"]:
        assert (
            len(
                {
                    run_result["validation_result"]["meta"][batch_key][field]
                    for run_result in result.run_results.values()
                }
            )
            == number_of_tables
        )

    # Check that every expectation result was correct.
    expected_results = taxi_benchmark_util.expected_validation_results()

    for run_result in result.run_results.values():
        actual_results = [
            result.to_json_dict()
            for result in run_result["validation_result"]["results"]
        ]
        assert len(expected_results) == len(actual_results)
        for expected_result, actual_result in zip(expected_results, actual_results):
            description_for_error_reporting = (
                f'{expected_result["expectation_config"]["expectation_type"]} result'
            )
            _recursively_assert_actual_result_matches_expected_result_keys(
                expected_result, actual_result, description_for_error_reporting
            )


def _recursively_assert_actual_result_matches_expected_result_keys(
    expected, actual, description_for_error_reporting
):
    """Assert that actual equals expected while ignoring key order and extra keys not present in expected.

    Expected mappings may be a subset of actual mappings -- this can be useful to make tests less fragile so that they
    don't incorrectly fail when new keys are added.

    Args:
        expected: The expected result. For mappings, every key in expected must be present in actual with the same value
            (while recursively ignoring nested key order and extra nested keys in any nested values).
        actual: The actual result for comparison with expected result.
        description_for_error_reporting: Description to provide context during error reporting. For each recursive call,
            the description is updated to also include the key. For example, if the initial description is
            "expect_table_columns_to_match_set result" and the assertion fails with the "raised_exception" key nested in
            the "exception_info" key then pytest will report with a description as shown in the following:

            >           assert expected == actual, description_for_error_reporting
            E           AssertionError: expect_table_columns_to_match_set result["exception_info"]["raised_exception"]
            E           assert True == False
            E             +True
            E             -False
    """
    if isinstance(expected, Mapping):
        for expected_key in expected.keys():
            assert expected_key in actual.keys(), description_for_error_reporting
            _recursively_assert_actual_result_matches_expected_result_keys(
                expected[expected_key],
                actual[expected_key],
                description_for_error_reporting + f'["{expected_key}"]',
            )
    else:
        assert expected == actual, description_for_error_reporting


def _skip_if_bigquery_performance_tests_not_enabled(
    pytestconfig: _pytest.config.Config,
):
    if not pytestconfig.getoption("bigquery") or not pytestconfig.getoption(
        "performance_tests"
    ):
        pytest.skip(
            "This test requires --bigquery and --performance-tests flags to run."
        )


if __name__ == "__main__":
    # For profiling, it can be useful to support running this script directly instead of using pytest to run.
    sys.exit(pytest.main(sys.argv))



from typing import Optional

from great_expectations.core.usage_statistics.anonymizers.base import BaseAnonymizer
from great_expectations.data_context.store.store_backend import StoreBackend


class StoreBackendAnonymizer(BaseAnonymizer):
    def __init__(
        self,
        aggregate_anonymizer: "Anonymizer",  # noqa: F821
        salt: Optional[str] = None,
    ) -> None:
        super().__init__(salt=salt)

        self._aggregate_anonymizer = aggregate_anonymizer

    def anonymize(
        self,
        obj: Optional[object] = None,
        store_backend_obj: Optional[StoreBackend] = None,
        store_backend_object_config: Optional[dict] = None,
    ) -> dict:
        assert (
            store_backend_obj or store_backend_object_config
        ), "Must pass store_backend_obj or store_backend_object_config."
        anonymized_info_dict = {}
        if store_backend_obj is not None:
            self._anonymize_object_info(
                object_=store_backend_obj,
                anonymized_info_dict=anonymized_info_dict,
            )
        else:
            class_name = store_backend_object_config.get("class_name")
            module_name = store_backend_object_config.get("module_name")
            if module_name is None:
                module_name = "great_expectations.data_context.store"
            self._anonymize_object_info(
                object_config={"class_name": class_name, "module_name": module_name},
                anonymized_info_dict=anonymized_info_dict,
            )
        return anonymized_info_dict

    def can_handle(self, obj: Optional[object] = None, **kwargs) -> bool:
        return (obj is not None and isinstance(obj, StoreBackend)) or (
            "store_backend_object_config" in kwargs
        )



import os
import shutil
from unittest import mock

import pytest
from click.testing import CliRunner, Result

from great_expectations.cli import cli
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.templates import CONFIG_VARIABLES_TEMPLATE
from great_expectations.util import gen_directory_tree_str, get_context
from tests.cli.test_cli import yaml
from tests.cli.utils import assert_no_logging_messages_or_tracebacks


@pytest.mark.parametrize(
    "invocation,input",
    [
        ("--v3-api init", "Y\n"),
        ("--v3-api --assume-yes init", ""),
    ],
)
@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_cli_init_on_new_project(
    mock_emit, invocation, input, caplog, tmp_path, monkeypatch
):
    monkeypatch.delenv(
        "GE_USAGE_STATS", raising=False
    )  # Undo the project-wide test default
    project_dir_path = tmp_path / "test_cli_init_diff"
    project_dir_path.mkdir()
    project_dir = str(project_dir_path)

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    result = runner.invoke(
        cli,
        invocation,
        input=input,
        catch_exceptions=False,
    )
    stdout = result.output

    assert len(stdout) < 6000, "CLI output is unreasonably long."
    assert "Always know what to expect from your data" in stdout
    lets_create_data_context = (
        "Let's create a new Data Context to hold your project configuration."
    )
    if invocation == "--v3-api init":
        assert lets_create_data_context in stdout
    if invocation == "--v3-api --assume-yes init":
        assert lets_create_data_context not in stdout
    assert (
        "Congratulations! You are now ready to customize your Great Expectations configuration."
        in stdout
    )
    assert (
        "You can customize your configuration in many ways. Here are some examples:"
        in stdout
    )

    assert os.path.isdir(os.path.join(project_dir, "great_expectations"))
    config_path = os.path.join(project_dir, "great_expectations/great_expectations.yml")
    assert os.path.isfile(config_path)

    config = yaml.load(open(config_path))
    assert config["datasources"] == {}

    obs_tree = gen_directory_tree_str(os.path.join(project_dir, "great_expectations"))

    assert (
        obs_tree
        == """great_expectations/
    .gitignore
    great_expectations.yml
    checkpoints/
    expectations/
        .ge_store_backend_id
    plugins/
        custom_data_docs/
            renderers/
            styles/
                data_docs_custom_styles.css
            views/
    profilers/
    uncommitted/
        config_variables.yml
        data_docs/
        validations/
            .ge_store_backend_id
"""
    )

    assert mock_emit.call_count == 2
    assert mock_emit.call_args_list == [
        mock.call(
            {"event_payload": {}, "event": "data_context.__init__", "success": True}
        ),
        mock.call(
            {
                "event_payload": {"api_version": "v3"},
                "event": "cli.init.create",
                "success": True,
            }
        ),
    ]

    # Assertions to make sure we didn't regress to v2 API
    assert "What data would you like Great Expectations to connect to" not in stdout
    assert "What are you processing your files with" not in stdout
    assert (
        "Enter the path of a data file (relative or absolute, s3a:// and gs:// paths are ok too)"
        not in stdout
    )
    assert "Name the new Expectation Suite [Titanic.warning]" not in stdout
    assert (
        "Great Expectations will choose a couple of columns and generate expectations about them"
        not in stdout
    )
    assert "Generating example Expectation Suite..." not in stdout
    assert "Building" not in stdout
    assert "Done generating example Expectation Suite" not in stdout

    assert_no_logging_messages_or_tracebacks(caplog, result)


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_cancelled_cli_init_on_new_project(mock_emit, caplog, tmp_path, monkeypatch):
    monkeypatch.delenv(
        "GE_USAGE_STATS", raising=False
    )  # Undo the project-wide test default
    project_dir_path = tmp_path / "test_cli_init_diff"
    project_dir_path.mkdir()
    project_dir = str(project_dir_path)

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    result = runner.invoke(
        cli,
        "--v3-api init",
        input="n\n",
        catch_exceptions=False,
    )
    stdout = result.output

    assert len(stdout) < 6000, "CLI output is unreasonably long."
    assert "Always know what to expect from your data" in stdout
    assert (
        "Let's create a new Data Context to hold your project configuration." in stdout
    )
    assert "OK. You must run" in stdout
    assert "great_expectations init" in stdout
    assert "to fix the missing files!" in stdout

    assert (
        "Congratulations! You are now ready to customize your Great Expectations configuration."
        not in stdout
    )
    assert (
        "You can customize your configuration in many ways. Here are some examples:"
        not in stdout
    )

    assert not os.path.isdir(os.path.join(project_dir, "great_expectations"))
    config_path = os.path.join(project_dir, "great_expectations/great_expectations.yml")
    assert not os.path.isfile(config_path)

    obs_tree = gen_directory_tree_str(os.path.join(project_dir, "great_expectations"))

    assert obs_tree == ""

    assert mock_emit.call_count == 0
    assert mock_emit.call_args_list == []

    assert_no_logging_messages_or_tracebacks(caplog, result)


def test_cli_init_on_existing_project_with_no_uncommitted_dirs_answering_no_then_yes_to_fixing_them(
    caplog,
    monkeypatch,
    tmp_path,
):
    """
    This test walks through the onboarding experience.

    The user just checked an existing project out of source control and does
    not yet have an uncommitted directory.
    """
    root_dir_path = tmp_path / "hiya"
    root_dir_path.mkdir()
    root_dir = str(root_dir_path)

    # Create a new project from scratch that we will use for the test in the next step

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(root_dir)
    result = runner.invoke(
        cli,
        ["--v3-api", "init"],
        input="Y\n",
        catch_exceptions=False,
    )
    stdout = result.output
    assert result.exit_code == 0

    assert (
        "Congratulations! You are now ready to customize your Great Expectations configuration."
        in stdout
    )

    context = get_context(
        context_root_dir=os.path.join(root_dir, FileDataContext.GX_DIR)
    )
    uncommitted_dir = os.path.join(context.root_directory, "uncommitted")
    shutil.rmtree(uncommitted_dir)
    assert not os.path.isdir(uncommitted_dir)

    # Test the second invocation of init (answer N to update broken init)
    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(os.path.dirname(context.root_directory))
    with pytest.warns(UserWarning):
        result = runner.invoke(
            cli,
            ["--v3-api", "init"],
            input="N\n",
            catch_exceptions=False,
        )
    stdout = result.stdout

    assert result.exit_code == 0
    assert (
        "It looks like you have a partially initialized Great Expectations project. Would you like to fix this automatically by adding the following missing files (existing files will not be modified)?"
        in stdout
    )
    assert "Great Expectations added some missing files required to run." not in stdout
    assert "OK. You must run" in stdout
    assert "great_expectations init" in stdout
    assert "to fix the missing files!" in stdout
    assert "You may see new files in" not in stdout
    assert "Would you like to build & view this project's Data Docs!?" not in stdout

    assert not os.path.isdir(uncommitted_dir)
    config_var_path = os.path.join(uncommitted_dir, "config_variables.yml")
    assert not os.path.isfile(config_var_path)

    # Test the third invocation of init (answer Yes to update broken init)
    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(os.path.dirname(context.root_directory))
    with pytest.warns(
        UserWarning, match="Warning. An existing `great_expectations.yml` was found"
    ):
        result = runner.invoke(
            cli,
            ["--v3-api", "init"],
            input="Y\n",
            catch_exceptions=False,
        )
    stdout = result.stdout

    assert result.exit_code == 0
    assert (
        "It looks like you have a partially initialized Great Expectations project. Would you like to fix this automatically by adding the following missing files (existing files will not be modified)?"
        in stdout
    )
    assert "Great Expectations added some missing files required to run." in stdout
    assert "You may see new files in" in stdout

    assert "OK. You must run" not in stdout
    assert "great_expectations init" not in stdout
    assert "to fix the missing files!" not in stdout
    assert "Would you like to build & view this project's Data Docs!?" not in stdout

    assert os.path.isdir(uncommitted_dir)
    config_var_path = os.path.join(uncommitted_dir, "config_variables.yml")
    assert os.path.isfile(config_var_path)
    with open(config_var_path) as f:
        assert f.read() == CONFIG_VARIABLES_TEMPLATE

    assert_no_logging_messages_or_tracebacks(caplog, result)


def test_cli_init_on_complete_existing_project_all_uncommitted_dirs_exist(
    caplog,
    monkeypatch,
    tmp_path,
):
    root_dir_path = tmp_path / "hiya"
    root_dir_path.mkdir()
    root_dir = str(root_dir_path)

    # Create a new project from scratch that we will use for the test in the next step

    runner: CliRunner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(root_dir)
    result: Result = runner.invoke(
        cli,
        ["--v3-api", "init"],
        input="Y\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0

    # Now the test begins - rerun the init on an existing project

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(root_dir)
    with pytest.warns(
        UserWarning, match="Warning. An existing `great_expectations.yml` was found"
    ):
        result = runner.invoke(
            cli,
            ["--v3-api", "init"],
            input="",
            catch_exceptions=False,
        )
    stdout = result.stdout

    assert result.exit_code == 0
    assert "This looks like an existing project that" in stdout
    assert "appears complete" in stdout
    assert "ready to roll" in stdout
    assert "Would you like to build & view this project's Data Docs" not in stdout
    assert_no_logging_messages_or_tracebacks(caplog, result)
